#include "Cubemap.h"

Cubemap::Cubemap(VulkanDevice vulkanDevice, Pipeline* pipeline, Model* model, CubemapTexture* material, bool dynamic) {
    m_vulkanDevice = vulkanDevice;
    m_device = m_vulkanDevice.getVkDevice();
    m_model = model;
    m_pipeline = pipeline;
    m_material = material;
    m_dynamic = dynamic;

    createUniformBuffers();
    createDescriptorPool();
    createDescriptorSets();
};

VkBuffer& Cubemap::getVertexBuffer() { return m_model->getVertexBuffer(); }
VkBuffer& Cubemap::getIndexBuffer() { return m_model->getIndexBuffer(); }
VkDescriptorSet& Cubemap::getDescriptorSet(int frame) { return m_descriptorSets.at(frame); }
std::vector<uint32_t> Cubemap::getIndices() { return m_model->getIndices(); }
VkPipeline Cubemap::getPipeline() { return m_pipeline->getPipeline(); };
VkPipeline Cubemap::getDrawToCubePipeline() {
    if (!m_dynamic) {
        throw std::runtime_error("cannot return this pipeline on static cubemap!");
    }
    return m_drawToCubemapPipeline->getPipeline(); 
};
VkPipeline Cubemap::getDrawCubeToCubePipeline() {
    if (!m_dynamic) {
        throw std::runtime_error("cannot return this pipeline on static cubemap!");
    }
    return m_drawCubeToCubemapPipeline->getPipeline();
};
void Cubemap::setDrawToCubePipeline(ImagePipeline* pipeline) {
    if (!m_dynamic) {
        throw std::runtime_error("cannot set this pipeline on static cubemap!");
    }
    m_drawToCubemapPipeline = pipeline;
}
void Cubemap::setDrawCubeToCubePipeline(Pipeline* pipeline) {
    if (!m_dynamic) {
        throw std::runtime_error("cannot set this pipeline on static cubemap!");
    }
    m_drawCubeToCubemapPipeline = pipeline;
}
VkPipelineLayout Cubemap::getPipelineLayout() { return m_pipeline->getPipelineLayout(); };
VkPipelineLayout Cubemap::getDrawToCubePipelineLayout() { 
    if (!m_dynamic) {
        throw std::runtime_error("cannot return this pipeline layout on static cubemap!");
    }
    return m_drawToCubemapPipeline->getPipelineLayout();
};
VkPipelineLayout Cubemap::getDrawCubeToCubePipelineLayout() {
    if (!m_dynamic) {
        throw std::runtime_error("cannot return this pipeline layout on static cubemap!");
    }
    return m_drawCubeToCubemapPipeline->getPipelineLayout();
};
CubemapTexture* Cubemap::getTexture() { return m_material; };

void Cubemap::createUniformBuffers() {
    VkDeviceSize bufferSize = sizeof(UniformBufferObject);

    m_uniformBuffers.resize(MAX_FRAMES_IN_FLIGHT);
    m_uniformBuffersMemory.resize(MAX_FRAMES_IN_FLIGHT);
    m_uniformBuffersMapped.resize(MAX_FRAMES_IN_FLIGHT);

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        m_vulkanDevice.createBuffer(bufferSize, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, m_uniformBuffers[i], m_uniformBuffersMemory[i]);

        vkMapMemory(m_device, m_uniformBuffersMemory[i], 0, bufferSize, 0, &m_uniformBuffersMapped[i]);
    }

    if (m_dynamic) {
        VkDeviceSize cameraBufferSize = sizeof(CameraInfoUniformBufferObject);

        m_cameraUniformBuffers.resize(MAX_FRAMES_IN_FLIGHT);
        m_cameraUniformBuffersMemory.resize(MAX_FRAMES_IN_FLIGHT);
        m_cameraUniformBuffersMapped.resize(MAX_FRAMES_IN_FLIGHT);

        for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
            m_vulkanDevice.createBuffer(cameraBufferSize, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, m_cameraUniformBuffers[i], m_cameraUniformBuffersMemory[i]);

            vkMapMemory(m_device, m_cameraUniformBuffersMemory[i], 0, cameraBufferSize, 0, &m_cameraUniformBuffersMapped[i]);
        }
    }
}

void Cubemap::createDescriptorPool() {

    std::vector<VkDescriptorPoolSize> poolSizes;

    poolSizes.push_back({ VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT) });
    poolSizes.push_back({ VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT) });

    if (m_dynamic) {
        poolSizes.push_back({ VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT) });
        poolSizes.push_back({ VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT) });
    }

    VkDescriptorPoolCreateInfo poolInfo{};
    poolInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
    poolInfo.poolSizeCount = static_cast<uint32_t>(poolSizes.size());
    poolInfo.pPoolSizes = poolSizes.data();
    poolInfo.maxSets = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    if (vkCreateDescriptorPool(m_device, &poolInfo, nullptr, &m_descriptorPool) != VK_SUCCESS) {
        throw std::runtime_error("failed to create descriptor pool!");
    }
}

void Cubemap::createDescriptorSets() {
    std::vector<VkDescriptorSetLayout> layouts(MAX_FRAMES_IN_FLIGHT, m_pipeline->getDescriptorSetLayout());
    VkDescriptorSetAllocateInfo allocInfo{};
    allocInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    allocInfo.descriptorPool = m_descriptorPool;
    allocInfo.descriptorSetCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);
    allocInfo.pSetLayouts = layouts.data();

    m_descriptorSets.resize(MAX_FRAMES_IN_FLIGHT);
    if (vkAllocateDescriptorSets(m_device, &allocInfo, m_descriptorSets.data()) != VK_SUCCESS) {
        throw std::runtime_error("failed to allocate descriptor sets!");
    }

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        VkDescriptorBufferInfo bufferInfo{};
        bufferInfo.buffer = m_uniformBuffers[i];
        bufferInfo.offset = 0;
        bufferInfo.range = sizeof(UniformBufferObject);

        VkDescriptorImageInfo imageInfo{};
        imageInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        imageInfo.imageView = m_material->getTextureImageView(i);
        imageInfo.sampler = m_material->getTextureSampler();

        std::vector<VkWriteDescriptorSet> descriptorWrites{};

        VkWriteDescriptorSet uboWrite{};
        uboWrite.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        uboWrite.dstSet = m_descriptorSets[i];
        uboWrite.dstBinding = 0;
        uboWrite.dstArrayElement = 0;
        uboWrite.descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
        uboWrite.descriptorCount = 1;
        uboWrite.pBufferInfo = &bufferInfo;
        descriptorWrites.push_back(uboWrite);

        VkWriteDescriptorSet samplerWrite{};
        samplerWrite.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        samplerWrite.dstSet = m_descriptorSets[i];
        samplerWrite.dstBinding = 1;
        samplerWrite.dstArrayElement = 0;
        samplerWrite.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        samplerWrite.descriptorCount = 1;
        samplerWrite.pImageInfo = &imageInfo;
        descriptorWrites.push_back(samplerWrite);    
        
        if (m_dynamic) {    
            VkDescriptorBufferInfo cameraBuffer{};
            cameraBuffer.buffer = m_cameraUniformBuffers[i];
            cameraBuffer.offset = 0;
            cameraBuffer.range = sizeof(CameraInfoUniformBufferObject);

            DynEnvironmentTexture* dynTexture = dynamic_cast<DynEnvironmentTexture*>(m_material);
            VkDescriptorImageInfo depthInfo{};
            depthInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
            depthInfo.imageView = dynTexture->getDepthImageView(i);
            depthInfo.sampler = dynTexture->getTextureSampler();

            VkWriteDescriptorSet cameraWrite{};
            cameraWrite.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
            cameraWrite.dstSet = m_descriptorSets[i];
            cameraWrite.dstBinding = 2;
            cameraWrite.dstArrayElement = 0;
            cameraWrite.descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
            cameraWrite.descriptorCount = 1;
            cameraWrite.pBufferInfo = &cameraBuffer;
            descriptorWrites.push_back(cameraWrite);

            VkWriteDescriptorSet depthSamplerWrite{};
            depthSamplerWrite.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
            depthSamplerWrite.dstSet = m_descriptorSets[i];
            depthSamplerWrite.dstBinding = 3;
            depthSamplerWrite.dstArrayElement = 0;
            depthSamplerWrite.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
            depthSamplerWrite.descriptorCount = 1;
            depthSamplerWrite.pImageInfo = &depthInfo;
            descriptorWrites.push_back(depthSamplerWrite);
        }

        vkUpdateDescriptorSets(m_device, static_cast<uint32_t>(descriptorWrites.size()), descriptorWrites.data(), 0, nullptr);
    }
}

void Cubemap::setPosition(glm::vec3 position) {
    if (!m_dynamic) {
        throw std::runtime_error("cannot set position on static cubemap!");
    }
    m_position = position;
};


glm::vec3 Cubemap::getPosition() {
    if (!m_dynamic) {
        throw std::runtime_error("cannot get position of static cubemap!");
    }
    return m_position;
};

void Cubemap::update(uint32_t currentImage, Camera& camera) {
    // TODO: calculate new path position, orientation and corresponding model matrix

    UniformBufferObject ubo{};
    ubo.model = glm::mat4(1.0f);
    ubo.view = glm::mat4(glm::mat3(camera.getViewMat()));
    ubo.proj = camera.getProjMat();

    memcpy(m_uniformBuffersMapped[currentImage], &ubo, sizeof(ubo));

    if (m_dynamic) {
        CameraInfoUniformBufferObject cubo{};
        cubo.cameraWorldPos = glm::vec4(camera.getPosition(), 1.0f);

        memcpy(m_cameraUniformBuffersMapped[currentImage], &cubo, sizeof(cubo));
    }
}

void Cubemap::updateForCubemapDraw(uint32_t currentImage, glm::mat4 viewMat, glm::mat4 projMat) {
    UniformBufferObject ubo{};
    ubo.model = glm::mat4(1.0f);
    ubo.view = glm::mat4(glm::mat3(viewMat));
    ubo.proj = projMat;

    memcpy(m_uniformBuffersMapped[currentImage], &ubo, sizeof(ubo));
};

void Cubemap::updateWithTRS(uint32_t currentImage, Camera& camera, glm::vec3 translation, float angle, glm::vec3 rotation, glm::vec3 scale) {
    UniformBufferObject ubo{};
    translation += m_position;
    //ubo.model = glm::translate(glm::rotate(glm::scale(glm::mat4(1.0f), scale), glm::radians(angle), rotation), translation);
    //ubo.model = glm::translate(glm::scale(glm::mat4(1.0f), scale), translation);
    ubo.model = glm::scale(glm::translate(glm::mat4(1.0f), getPosition()), scale);
    ubo.view = camera.getViewMat();
    ubo.proj = camera.getProjMat();

    memcpy(m_uniformBuffersMapped[currentImage], &ubo, sizeof(ubo));

    if (m_dynamic) {
        CameraInfoUniformBufferObject cubo{};
        cubo.cameraWorldPos = glm::vec4(camera.getPosition(), 1.0f);

        memcpy(m_cameraUniformBuffersMapped[currentImage], &cubo, sizeof(cubo));
    }
};

void Cubemap::cleanup() {
    m_material->cleanup();
    m_model->cleanup();

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        vkDestroyBuffer(m_device, m_uniformBuffers[i], nullptr);
        vkFreeMemory(m_device, m_uniformBuffersMemory[i], nullptr);

        if (m_dynamic) {
            vkDestroyBuffer(m_device, m_cameraUniformBuffers[i], nullptr);
            vkFreeMemory(m_device, m_cameraUniformBuffersMemory[i], nullptr);
        }
    }
    vkDestroyDescriptorPool(m_device, m_descriptorPool, nullptr);
}