#include "Material.h"

#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>

ImageTexture::ImageTexture(VulkanDevice vulkanDevice, std::string texturePath, std::string normalMapPath, glm::vec3 coefficients /*x = ambient, y = diffuse, z = specular*/, float metallic, float roughness, float specularAlpha) {
    m_vulkanDevice = vulkanDevice;
    m_device = m_vulkanDevice.getVkDevice();
    m_coefficients = coefficients;
    m_metallic = metallic;
    m_roughness = roughness;
    m_specularAlpha = specularAlpha;

    loadImage(texturePath);
    createTextureImage();
    createTextureImageView();
    createTextureSampler(&m_textureSampler);

    loadNormalMap(normalMapPath);
}

void ImageTexture::loadNormalMap(std::string path) {
    // Load normal map image
    int width, height, channels;
    stbi_uc* normalPixels = stbi_load(path.c_str(), &width, &height, &channels, STBI_rgb_alpha);
    VkDeviceSize imageSize = width * height * 4;

    if (!normalPixels) {
        throw std::runtime_error("failed to load normal map!");
    }

    // Create staging buffer
    VkBuffer stagingBuffer;
    VkDeviceMemory stagingBufferMemory;
    m_vulkanDevice.createBuffer(imageSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
        VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
        stagingBuffer, stagingBufferMemory);

    void* data;
    vkMapMemory(m_device, stagingBufferMemory, 0, imageSize, 0, &data);
    memcpy(data, normalPixels, static_cast<size_t>(imageSize));
    vkUnmapMemory(m_device, stagingBufferMemory);

    stbi_image_free(normalPixels);

    // Create normal map image
    createImage(width, height, VK_FORMAT_R8G8B8A8_SRGB, VK_IMAGE_TILING_OPTIMAL,
        VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT,
        VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, m_normalMap, m_normalMapMemory);

    // Setup image memory barrier
    transitionImageLayout(m_normalMap, VK_FORMAT_R8G8B8A8_SRGB,
        VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);

    // Copy buffer to image
    m_vulkanDevice.copyBufferToImage(stagingBuffer, m_normalMap,
        static_cast<uint32_t>(width), static_cast<uint32_t>(height),1);

    // Transition to shader reading
    transitionImageLayout(m_normalMap, VK_FORMAT_R8G8B8A8_SRGB,
        VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL);

    // Create view and sampler
    m_normalMapView = createImageView(m_normalMap, VK_FORMAT_R8G8B8A8_SRGB, VK_IMAGE_ASPECT_COLOR_BIT);
    createTextureSampler(&m_normalMapSampler);

    // Cleanup staging buffer
    vkDestroyBuffer(m_device, stagingBuffer, nullptr);
    vkFreeMemory(m_device, stagingBufferMemory, nullptr);
}

void ImageTexture::loadImage(std::string path) {
    m_pixels = stbi_load(path.c_str(), &m_texWidth, &m_texHeight, &m_texChannels, STBI_rgb_alpha);
    m_imageSize = m_texWidth * m_texHeight * 4;

    if (!m_pixels) {
        throw std::runtime_error("failed to load texture image!");
    }
}

void ImageTexture::transitionImageLayout(VkImage image, VkFormat format, VkImageLayout oldLayout, VkImageLayout newLayout) {
    VkCommandBuffer commandBuffer = m_vulkanDevice.beginSingleTimeCommands();

    VkImageMemoryBarrier barrier{};
    barrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
    barrier.oldLayout = oldLayout;
    barrier.newLayout = newLayout;
    barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    barrier.image = image;
    barrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    barrier.subresourceRange.baseMipLevel = 0;
    barrier.subresourceRange.levelCount = 1;
    barrier.subresourceRange.baseArrayLayer = 0;
    barrier.subresourceRange.layerCount = 1;

    VkPipelineStageFlags sourceStage;
    VkPipelineStageFlags destinationStage;

    if (oldLayout == VK_IMAGE_LAYOUT_UNDEFINED && newLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL) {
        barrier.srcAccessMask = 0;
        barrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;

        sourceStage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
        destinationStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
    }
    else if (oldLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL && newLayout == VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL) {
        barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        barrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;

        sourceStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
        destinationStage = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT;
    }
    else {
        throw std::invalid_argument("unsupported layout transition!");
    }

    vkCmdPipelineBarrier(
        commandBuffer,
        sourceStage, destinationStage,
        0,
        0, nullptr,
        0, nullptr,
        1, &barrier
    );

    m_vulkanDevice.endSingleTimeCommands(commandBuffer);
}

void StaticEnvironmentTexture::transitionImageLayout(VkImage image, VkFormat format, VkImageLayout oldLayout, VkImageLayout newLayout) {
    VkCommandBuffer commandBuffer = m_vulkanDevice.beginSingleTimeCommands();

    VkImageMemoryBarrier barrier{};
    barrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
    barrier.oldLayout = oldLayout;
    barrier.newLayout = newLayout;
    barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    barrier.image = image;
    barrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    barrier.subresourceRange.baseMipLevel = 0;
    barrier.subresourceRange.levelCount = 1;
    barrier.subresourceRange.baseArrayLayer = 0;
    barrier.subresourceRange.layerCount = NUMBER_OF_CUBEMAP_IMAGES;

    VkPipelineStageFlags sourceStage;
    VkPipelineStageFlags destinationStage;

    if (oldLayout == VK_IMAGE_LAYOUT_UNDEFINED && newLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL) {
        barrier.srcAccessMask = 0;
        barrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;

        sourceStage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
        destinationStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
    }
    else if (oldLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL && newLayout == VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL) {
        barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        barrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;

        sourceStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
        destinationStage = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT;
    }
    else {
        throw std::invalid_argument("unsupported layout transition!");
    }

    vkCmdPipelineBarrier(
        commandBuffer,
        sourceStage, destinationStage,
        0,
        0, nullptr,
        0, nullptr,
        1, &barrier
    );

    m_vulkanDevice.endSingleTimeCommands(commandBuffer);
}

void ImageTexture::createTextureImage() {
    VkBuffer stagingBuffer;
    VkDeviceMemory stagingBufferMemory;
    m_vulkanDevice.createBuffer(m_imageSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, stagingBuffer, stagingBufferMemory);

    void* data;
    vkMapMemory(m_device, stagingBufferMemory, 0, m_imageSize, 0, &data);
    memcpy(data, m_pixels, static_cast<size_t>(m_imageSize));
    vkUnmapMemory(m_device, stagingBufferMemory);

    stbi_image_free(m_pixels);

    createImage(m_texWidth, m_texHeight, VK_FORMAT_R8G8B8A8_SRGB, VK_IMAGE_TILING_OPTIMAL, VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, m_textureImage, m_textureImageMemory);

    transitionImageLayout(m_textureImage, VK_FORMAT_R8G8B8A8_SRGB, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);
    m_vulkanDevice.copyBufferToImage(stagingBuffer, m_textureImage, static_cast<uint32_t>(m_texWidth), static_cast<uint32_t>(m_texHeight), 1);
    transitionImageLayout(m_textureImage, VK_FORMAT_R8G8B8A8_SRGB, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL);

    vkDestroyBuffer(m_device, stagingBuffer, nullptr);
    vkFreeMemory(m_device, stagingBufferMemory, nullptr);
}

void ImageTexture::createImage(uint32_t width, uint32_t height, VkFormat format, VkImageTiling tiling, VkImageUsageFlags usage, VkMemoryPropertyFlags properties, VkImage& image, VkDeviceMemory& imageMemory) {
    VkImageCreateInfo imageInfo{};
    imageInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
    imageInfo.imageType = VK_IMAGE_TYPE_2D;
    imageInfo.extent.width = width;
    imageInfo.extent.height = height;
    imageInfo.extent.depth = 1;
    imageInfo.mipLevels = 1;
    imageInfo.arrayLayers = 1;
    imageInfo.format = format;
    imageInfo.tiling = tiling;
    imageInfo.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
    imageInfo.usage = usage;
    imageInfo.samples = VK_SAMPLE_COUNT_1_BIT;
    imageInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

    if (vkCreateImage(m_device, &imageInfo, nullptr, &image) != VK_SUCCESS) {
        throw std::runtime_error("failed to create image!");
    }

    VkMemoryRequirements memRequirements;
    vkGetImageMemoryRequirements(m_device, image, &memRequirements);

    VkMemoryAllocateInfo allocInfo{};
    allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    allocInfo.allocationSize = memRequirements.size;
    allocInfo.memoryTypeIndex = m_vulkanDevice.findMemoryType(memRequirements.memoryTypeBits, properties);

    if (vkAllocateMemory(m_device, &allocInfo, nullptr, &imageMemory) != VK_SUCCESS) {
        throw std::runtime_error("failed to allocate image memory!");
    }

    vkBindImageMemory(m_device, image, imageMemory, 0);
}

void StaticEnvironmentTexture::createImage(uint32_t width, uint32_t height, VkFormat format, VkImageTiling tiling, VkImageUsageFlags usage, VkMemoryPropertyFlags properties, VkImage& image, VkDeviceMemory& imageMemory) {
    VkImageCreateInfo imageInfo{};
    imageInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
    imageInfo.flags = VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT;
    imageInfo.imageType = VK_IMAGE_TYPE_2D;
    imageInfo.extent.width = width;
    imageInfo.extent.height = height;
    imageInfo.extent.depth = 1;
    imageInfo.mipLevels = 1;
    imageInfo.arrayLayers = NUMBER_OF_CUBEMAP_IMAGES;
    imageInfo.format = format;
    imageInfo.tiling = tiling;
    imageInfo.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
    imageInfo.usage = usage;
    imageInfo.samples = VK_SAMPLE_COUNT_1_BIT;
    imageInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

    if (vkCreateImage(m_device, &imageInfo, nullptr, &image) != VK_SUCCESS) {
        throw std::runtime_error("failed to create image!");
    }

    VkMemoryRequirements memRequirements;
    vkGetImageMemoryRequirements(m_device, image, &memRequirements);

    VkMemoryAllocateInfo allocInfo{};
    allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    allocInfo.allocationSize = memRequirements.size;
    allocInfo.memoryTypeIndex = m_vulkanDevice.findMemoryType(memRequirements.memoryTypeBits, properties);

    if (vkAllocateMemory(m_device, &allocInfo, nullptr, &imageMemory) != VK_SUCCESS) {
        throw std::runtime_error("failed to allocate image memory!");
    }

    vkBindImageMemory(m_device, image, imageMemory, 0);
}

void ImageTexture::createTextureImageView() {
    m_textureImageView = createImageView(m_textureImage, VK_FORMAT_R8G8B8A8_SRGB, VK_IMAGE_ASPECT_COLOR_BIT);
}

VkImageView& ImageTexture::getTextureImageView() {
    return m_textureImageView;
};

VkSampler& ImageTexture::getTextureSampler() {
    return m_textureSampler;
};

VkImageView& StaticEnvironmentTexture::getTextureImageView(int index) {
    return m_textureImageView;
};

VkSampler& StaticEnvironmentTexture::getTextureSampler() {
    return m_textureSampler;
};

VkImageView ImageTexture::createImageView(VkImage image, VkFormat format, VkImageAspectFlags aspectFlags) {
    VkImageViewCreateInfo viewInfo{};
    viewInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    viewInfo.image = image;
    viewInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
    viewInfo.format = format;
    viewInfo.subresourceRange.aspectMask = aspectFlags;
    viewInfo.subresourceRange.baseMipLevel = 0;
    viewInfo.subresourceRange.levelCount = 1;
    viewInfo.subresourceRange.baseArrayLayer = 0;
    viewInfo.subresourceRange.layerCount = 1;

    VkImageView imageView;
    if (vkCreateImageView(m_device, &viewInfo, nullptr, &imageView) != VK_SUCCESS) {
        throw std::runtime_error("failed to create texture image view!");
    }

    return imageView;
}

void ImageTexture::createTextureSampler(VkSampler* sampler) {
    VkPhysicalDeviceProperties properties{};
    vkGetPhysicalDeviceProperties(m_vulkanDevice.getVkPhysicalDevice(), &properties);

    VkSamplerCreateInfo samplerInfo{};
    samplerInfo.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerInfo.magFilter = VK_FILTER_LINEAR;
    samplerInfo.minFilter = VK_FILTER_LINEAR;
    samplerInfo.addressModeU = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerInfo.addressModeV = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerInfo.addressModeW = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerInfo.anisotropyEnable = VK_TRUE;
    samplerInfo.maxAnisotropy = properties.limits.maxSamplerAnisotropy;
    samplerInfo.borderColor = VK_BORDER_COLOR_INT_OPAQUE_BLACK;
    samplerInfo.unnormalizedCoordinates = VK_FALSE;
    samplerInfo.compareEnable = VK_FALSE;
    samplerInfo.compareOp = VK_COMPARE_OP_ALWAYS;
    samplerInfo.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;

    if (vkCreateSampler(m_device, &samplerInfo, nullptr, sampler) != VK_SUCCESS) {
        throw std::runtime_error("failed to create texture sampler!");
    }
}

void ImageTexture::cleanup() {
    if (!clean) {
        vkDestroySampler(m_device, m_textureSampler, nullptr);
        vkDestroyImageView(m_device, m_textureImageView, nullptr);

        vkDestroyImage(m_device, m_textureImage, nullptr);
        vkFreeMemory(m_device, m_textureImageMemory, nullptr);

        vkDestroySampler(m_device, m_normalMapSampler, nullptr);
        vkDestroyImageView(m_device, m_normalMapView, nullptr);
        vkDestroyImage(m_device, m_normalMap, nullptr);
        vkFreeMemory(m_device, m_normalMapMemory, nullptr);
        clean = true;
    }
}

void StaticEnvironmentTexture::cleanup() {
    if (!clean) {
        vkDestroySampler(m_device, m_textureSampler, nullptr);
        vkDestroyImageView(m_device, m_textureImageView, nullptr);

        vkDestroyImage(m_device, m_textureImage, nullptr);
        vkFreeMemory(m_device, m_textureImageMemory, nullptr);
        clean = true;
    }
}

StaticEnvironmentTexture::StaticEnvironmentTexture(VulkanDevice vulkanDevice, std::string path) {
    m_vulkanDevice = vulkanDevice;
    m_device = m_vulkanDevice.getVkDevice();

    loadImage(path);
    create();
}

void StaticEnvironmentTexture::loadImage(std::string path) {
    
    for (int i = 0; i < NUMBER_OF_CUBEMAP_IMAGES; i++) {
        std::string temp = path.append("/");

        switch (i) {
        case POSX:
            temp.append("posx.jpg");
            break;
        case NEGX:
            temp.append("negx.jpg");
            break;
        case POSY:
            temp.append("posy.jpg");
            break;
        case NEGY:
            temp.append("negy.jpg");
            break;
        case POSZ:
            temp.append("posz.jpg");
            break;
        case NEGZ:
            temp.append("negz.jpg");
            break;
        default:
            break;
        }

        m_pixels[i] = stbi_load(temp.c_str(), &m_texWidth, &m_texHeight, &m_texChannels, STBI_rgb_alpha);

        if (!m_pixels[i]) {
            std::string err = "Cubemap: Couldn't load file ";
            err.append(temp);
            throw std::runtime_error(err);
        }
    }
}

void StaticEnvironmentTexture::create() {
    VkDeviceSize layerSize = m_texWidth * m_texHeight * 4;
    m_imageSize = layerSize * NUMBER_OF_CUBEMAP_IMAGES;

    VkBuffer stagingBuffer;
    VkDeviceMemory stagingBufferMemory;

    m_vulkanDevice.createBuffer(m_imageSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
        VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, stagingBuffer, stagingBufferMemory);

    void* data;
    vkMapMemory(m_device, stagingBufferMemory, 0, m_imageSize, 0, &data);

    for (int i = 0; i < NUMBER_OF_CUBEMAP_IMAGES; i++) {
        uint64_t memAddress = reinterpret_cast<uint64_t>(data) + (layerSize * i);
        memcpy(reinterpret_cast<void*>(memAddress), static_cast<void*>(m_pixels[i]), static_cast<size_t>(layerSize));
    }
   
    vkUnmapMemory(m_device, stagingBufferMemory);

    VkFormat format = VK_FORMAT_R8G8B8A8_SRGB;

    // Image

    createImage(m_texWidth, m_texHeight, format,                                                       
        VK_IMAGE_TILING_OPTIMAL,                                      
        VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT,
        VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, m_textureImage, m_textureImageMemory                           
    );

    transitionImageLayout(m_textureImage, format, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);

    m_vulkanDevice.copyBufferToImage(stagingBuffer,       
        m_textureImage,             
        static_cast<uint32_t>(m_texWidth),  
        static_cast<uint32_t>(m_texHeight), 
        NUMBER_OF_CUBEMAP_IMAGES     
    );

    transitionImageLayout(m_textureImage, format, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL);

    vkDestroyBuffer(m_device, stagingBuffer, nullptr);
    vkFreeMemory(m_device, stagingBufferMemory, nullptr);


    // Sampler

    VkSamplerCreateInfo samplerCreateInfo{};
    samplerCreateInfo.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerCreateInfo.magFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.minFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.addressModeU = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerCreateInfo.addressModeV = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerCreateInfo.addressModeW = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerCreateInfo.compareOp = VK_COMPARE_OP_NEVER;
    samplerCreateInfo.mipLodBias = 0.0f;
    samplerCreateInfo.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;
    samplerCreateInfo.minLod = 0.0f;
    samplerCreateInfo.maxLod = static_cast<float>(1);
    samplerCreateInfo.maxAnisotropy = 4.0;
    samplerCreateInfo.anisotropyEnable = VK_TRUE;
    samplerCreateInfo.borderColor = VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE;

    if (vkCreateSampler(m_device, &samplerCreateInfo, nullptr, &m_textureSampler) != VK_SUCCESS)
    {
        throw std::runtime_error("failed to create texture sampler!");
    }

    // Image view

    VkImageViewCreateInfo view{};
    view.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    view.viewType = VK_IMAGE_VIEW_TYPE_CUBE;
    view.format = format;
    view.components = { VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_G, VK_COMPONENT_SWIZZLE_B, VK_COMPONENT_SWIZZLE_A };
    // A subresource range describes the set of mip levels (and array layers) that can be accessed through this image
    // view It's possible to create multiple image views for a single image referring to different (and/or overlapping)
    // ranges of the image
    view.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    view.subresourceRange.baseMipLevel = 0;
    view.subresourceRange.baseArrayLayer = 0;
    view.subresourceRange.layerCount = NUMBER_OF_CUBEMAP_IMAGES;
    // Linear tiling usually won't support mip maps
    // Only set mip map count if optimal tiling is used
    view.subresourceRange.levelCount = 1;
    view.image = m_textureImage;

    if (vkCreateImageView(m_device, &view, nullptr, &m_textureImageView) != VK_SUCCESS)
    {
        throw std::runtime_error("failed to create image view!");
    }
}

DynEnvironmentTexture::DynEnvironmentTexture(VulkanDevice vulkanDevice, uint32_t width, uint32_t height, uint32_t channels) {
    m_vulkanDevice = vulkanDevice;
    m_device = vulkanDevice.getVkDevice();
    m_texWidth = width;
    m_texHeight = height;
    m_texChannels = channels;

    create();
}

void DynEnvironmentTexture::create() {
    VkDeviceSize layerSize = m_texWidth * m_texHeight * 4;
    m_imageSize = layerSize * NUMBER_OF_CUBEMAP_IMAGES;  

    VkFormat format = VK_FORMAT_R8G8B8A8_SRGB;
    VkFormat depthFormat = m_vulkanDevice.findDepthFormat();

    for (int i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        m_textureImages.resize(MAX_FRAMES_IN_FLIGHT);
        m_textureImageMemories.resize(MAX_FRAMES_IN_FLIGHT);
        m_cubemapImageViews.resize(MAX_FRAMES_IN_FLIGHT);
        m_colorAttachmentViews.resize(MAX_FRAMES_IN_FLIGHT * 6);

        VkBuffer stagingBuffer;
        VkDeviceMemory stagingBufferMemory;

        m_vulkanDevice.createBuffer(m_imageSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
            VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, stagingBuffer, stagingBufferMemory);

        // Image

        createImage(m_texWidth, m_texHeight, format,
            VK_IMAGE_TILING_OPTIMAL,
            VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
            VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, m_textureImages[i], m_textureImageMemories[i]
        );

        transitionImageLayout(m_textureImages[i], format, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_ASPECT_COLOR_BIT);

        m_vulkanDevice.copyBufferToImage(stagingBuffer,
            m_textureImages[i],
            static_cast<uint32_t>(m_texWidth),
            static_cast<uint32_t>(m_texHeight),
            NUMBER_OF_CUBEMAP_IMAGES
        );

        transitionImageLayout(m_textureImages[i], format, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, VK_IMAGE_ASPECT_COLOR_BIT);

        vkDestroyBuffer(m_device, stagingBuffer, nullptr);
        vkFreeMemory(m_device, stagingBufferMemory, nullptr);

        // Cubemap Image view

        VkImageViewCreateInfo view{};
        view.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
        view.viewType = VK_IMAGE_VIEW_TYPE_CUBE;
        view.format = format;
        view.components = { VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_G, VK_COMPONENT_SWIZZLE_B, VK_COMPONENT_SWIZZLE_A };
        view.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        view.subresourceRange.baseMipLevel = 0;
        view.subresourceRange.baseArrayLayer = 0;
        view.subresourceRange.layerCount = NUMBER_OF_CUBEMAP_IMAGES;
        view.subresourceRange.levelCount = 1;
        view.image = m_textureImages[i];

        if (vkCreateImageView(m_device, &view, nullptr, &m_cubemapImageViews[i]) != VK_SUCCESS)
        {
            throw std::runtime_error("failed to create image view!");
        }

        // Attachment Image views

        for (int j = 0; j < 6; j++) {
            int index = j;
            if (i > 0) {
                index = 6 + j;
            } 

            VkImageViewCreateInfo viewInfo{};
            viewInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            viewInfo.image = m_textureImages[i];
            viewInfo.viewType = VK_IMAGE_VIEW_TYPE_2D; 
            viewInfo.format = format;
            viewInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            viewInfo.subresourceRange.baseMipLevel = 0;  
            viewInfo.subresourceRange.levelCount = 1;
            viewInfo.subresourceRange.baseArrayLayer = j;  
            viewInfo.subresourceRange.layerCount = 1;  

            vkCreateImageView(m_device, &viewInfo, nullptr, &m_colorAttachmentViews[index]);
        }

        // Depth

        m_depthImages.resize(MAX_FRAMES_IN_FLIGHT);
        m_depthImageMemories.resize(MAX_FRAMES_IN_FLIGHT);
        m_depthCubemapImageViews.resize(MAX_FRAMES_IN_FLIGHT);
        m_depthAttachmentViews.resize(MAX_FRAMES_IN_FLIGHT * 6);

        VkBuffer depthStagingBuffer;
        VkDeviceMemory depthStagingBufferMemory;

        m_vulkanDevice.createBuffer(m_imageSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
            VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, depthStagingBuffer, depthStagingBufferMemory);

        // Image

        createImage(m_texWidth, m_texHeight, depthFormat,
            VK_IMAGE_TILING_OPTIMAL,
            VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT,
            VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, m_depthImages[i], m_depthImageMemories[i]
        );

        transitionImageLayout(m_depthImages[i], depthFormat, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_ASPECT_DEPTH_BIT);

        m_vulkanDevice.copyDepthBufferToImage(depthStagingBuffer,
            m_depthImages[i],
            static_cast<uint32_t>(m_texWidth),
            static_cast<uint32_t>(m_texHeight),
            NUMBER_OF_CUBEMAP_IMAGES
        );

        transitionImageLayout(m_depthImages[i], depthFormat, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, VK_IMAGE_ASPECT_DEPTH_BIT);

        vkDestroyBuffer(m_device, depthStagingBuffer, nullptr);
        vkFreeMemory(m_device, depthStagingBufferMemory, nullptr);

        // Cubemap Image view

        VkImageViewCreateInfo depthView{};
        depthView.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
        depthView.viewType = VK_IMAGE_VIEW_TYPE_CUBE;
        depthView.format = depthFormat;
        depthView.components = { VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_G, VK_COMPONENT_SWIZZLE_B, VK_COMPONENT_SWIZZLE_A };
        depthView.subresourceRange.aspectMask = VK_IMAGE_ASPECT_DEPTH_BIT;
        depthView.subresourceRange.baseMipLevel = 0;
        depthView.subresourceRange.baseArrayLayer = 0;
        depthView.subresourceRange.layerCount = NUMBER_OF_CUBEMAP_IMAGES;
        depthView.subresourceRange.levelCount = 1;
        depthView.image = m_depthImages[i];

        if (vkCreateImageView(m_device, &depthView, nullptr, &m_depthCubemapImageViews[i]) != VK_SUCCESS)
        {
            throw std::runtime_error("failed to create image view!");
        }

        // Attachment Image views

        for (int j = 0; j < 6; j++) {
            int index = j;
            if (i > 0) {
                index = 6 + j;
            }

            VkImageViewCreateInfo viewInfo{};
            viewInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            viewInfo.image = m_depthImages[i];
            viewInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            viewInfo.format = depthFormat;
            viewInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_DEPTH_BIT;
            viewInfo.subresourceRange.baseMipLevel = 0;
            viewInfo.subresourceRange.levelCount = 1;
            viewInfo.subresourceRange.baseArrayLayer = j;
            viewInfo.subresourceRange.layerCount = 1;

            vkCreateImageView(m_device, &viewInfo, nullptr, &m_depthAttachmentViews[index]);
        }
    }

    // Sampler

    VkSamplerCreateInfo samplerCreateInfo{};
    samplerCreateInfo.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerCreateInfo.magFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.minFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.addressModeU = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerCreateInfo.addressModeV = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerCreateInfo.addressModeW = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerCreateInfo.compareOp = VK_COMPARE_OP_NEVER;
    samplerCreateInfo.mipLodBias = 0.0f;
    samplerCreateInfo.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;
    samplerCreateInfo.minLod = 0.0f;
    samplerCreateInfo.maxLod = static_cast<float>(1);
    samplerCreateInfo.maxAnisotropy = 4.0;
    samplerCreateInfo.anisotropyEnable = VK_TRUE;
    samplerCreateInfo.borderColor = VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE;

    if (vkCreateSampler(m_device, &samplerCreateInfo, nullptr, &m_textureSampler) != VK_SUCCESS)
    {
        throw std::runtime_error("failed to create texture sampler!");
    }
}

void DynEnvironmentTexture::createImage(uint32_t width, uint32_t height, VkFormat format,
    VkImageTiling tiling, VkImageUsageFlags usage, VkMemoryPropertyFlags properties, VkImage& image, VkDeviceMemory& imageMemory) {
    VkImageCreateInfo imageInfo{};
    imageInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
    imageInfo.flags = VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT;
    imageInfo.imageType = VK_IMAGE_TYPE_2D;
    imageInfo.extent.width = width;
    imageInfo.extent.height = height;
    imageInfo.extent.depth = 1;
    imageInfo.mipLevels = 1;
    imageInfo.arrayLayers = NUMBER_OF_CUBEMAP_IMAGES;
    imageInfo.format = format;
    imageInfo.tiling = tiling;
    imageInfo.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
    imageInfo.usage = usage;
    imageInfo.samples = VK_SAMPLE_COUNT_1_BIT;
    imageInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

    if (vkCreateImage(m_device, &imageInfo, nullptr, &image) != VK_SUCCESS) {
        throw std::runtime_error("failed to create image!");
    }

    VkMemoryRequirements memRequirements;
    vkGetImageMemoryRequirements(m_device, image, &memRequirements);

    VkMemoryAllocateInfo allocInfo{};
    allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    allocInfo.allocationSize = memRequirements.size;
    allocInfo.memoryTypeIndex = m_vulkanDevice.findMemoryType(memRequirements.memoryTypeBits, properties);

    if (vkAllocateMemory(m_device, &allocInfo, nullptr, &imageMemory) != VK_SUCCESS) {
        throw std::runtime_error("failed to allocate image memory!");
    }

    vkBindImageMemory(m_device, image, imageMemory, 0);
}

VkImageView& DynEnvironmentTexture::getTextureImageView(int index) {
    return m_cubemapImageViews[index];
};

VkImageView& DynEnvironmentTexture::getDepthImageView(int index) {
    return m_depthCubemapImageViews[index];
}

VkSampler& DynEnvironmentTexture::getTextureSampler() {
    return m_textureSampler;
};

void DynEnvironmentTexture::cleanup() {
    if (!clean) {
        vkDestroySampler(m_device, m_textureSampler, nullptr);

        for (int i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
            vkDestroyImageView(m_device, m_cubemapImageViews[i], nullptr);
            vkDestroyImage(m_device, m_textureImages[i], nullptr);
            vkFreeMemory(m_device, m_textureImageMemories[i], nullptr);

            vkDestroyImageView(m_device, m_depthCubemapImageViews[i], nullptr);
            vkDestroyImage(m_device, m_depthImages[i], nullptr);
            vkFreeMemory(m_device, m_depthImageMemories[i], nullptr);

            for (int j = i * 6; j < i * 6 + 6; j++) {
                vkDestroyImageView(m_device, m_colorAttachmentViews[j], nullptr);
                vkDestroyImageView(m_device, m_depthAttachmentViews[j], nullptr);
            }     
        }
        
        clean = true;
    }
};

std::vector<VkImageView> DynEnvironmentTexture::getAttachmentViews() {
    return m_colorAttachmentViews;
}

std::vector<VkImageView> DynEnvironmentTexture::getDepthAttachmentViews() {
    return m_depthAttachmentViews;
}


uint32_t DynEnvironmentTexture::getWidth() { return m_texWidth; }
uint32_t DynEnvironmentTexture::getHeight() { return m_texHeight; }

VkImage& DynEnvironmentTexture::getTextureImage(int index) {
    return m_textureImages[index];
};

void DynEnvironmentTexture::transitionImageLayout(VkImage image, VkFormat format, VkImageLayout oldLayout, VkImageLayout newLayout, VkImageAspectFlagBits aspectFlags) {
    VkCommandBuffer commandBuffer = m_vulkanDevice.beginSingleTimeCommands();

    VkImageMemoryBarrier barrier{};
    barrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
    barrier.oldLayout = oldLayout;
    barrier.newLayout = newLayout;
    barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    barrier.image = image;
    barrier.subresourceRange.aspectMask = aspectFlags;
    barrier.subresourceRange.baseMipLevel = 0;
    barrier.subresourceRange.levelCount = 1;
    barrier.subresourceRange.baseArrayLayer = 0;
    barrier.subresourceRange.layerCount = NUMBER_OF_CUBEMAP_IMAGES;

    VkPipelineStageFlags sourceStage;
    VkPipelineStageFlags destinationStage;

    if (oldLayout == VK_IMAGE_LAYOUT_UNDEFINED && newLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL) {
        barrier.srcAccessMask = 0;
        barrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;

        sourceStage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
        destinationStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
    }
    else if (oldLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL && newLayout == VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL) {
        barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        barrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;

        sourceStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
        destinationStage = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT;
    }
    else {
        throw std::invalid_argument("unsupported layout transition!");
    }

    vkCmdPipelineBarrier(
        commandBuffer,
        sourceStage, destinationStage,
        0,
        0, nullptr,
        0, nullptr,
        1, &barrier
    );

    m_vulkanDevice.endSingleTimeCommands(commandBuffer);
}

DefferedTexture::DefferedTexture(VulkanDevice vulkanDevice) {
    m_vulkanDevice = vulkanDevice;
    m_device = vulkanDevice.getVkDevice();
    
    create();
}

void DefferedTexture::create() {
    VkFormat format = VK_FORMAT_R8G8B8A8_UNORM;
    VkFormat normalFormat = VK_FORMAT_R16G16B16A16_SFLOAT;
    VkFormat worldPosFormat = VK_FORMAT_R16G16B16A16_SFLOAT;

    // Images and Views
    createForType(m_albedoImages, m_albedoMemories, m_albedoViews, format);
    createForType(m_normalImages, m_normalMemories, m_normalViews, normalFormat);
    createForType(m_coefficientImages, m_coefficientMemories, m_coefficientViews, format);
    createForType(m_materialImages, m_materialMemories, m_materialViews, format);
    createForType(m_worldPosImages, m_worldPosMemories, m_worldPosViews, worldPosFormat);

    // Sampler
    VkSamplerCreateInfo samplerCreateInfo{};
    samplerCreateInfo.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerCreateInfo.magFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.minFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.addressModeU = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerCreateInfo.addressModeV = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerCreateInfo.addressModeW = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    samplerCreateInfo.compareOp = VK_COMPARE_OP_NEVER;
    samplerCreateInfo.mipLodBias = 0.0f;
    samplerCreateInfo.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;
    samplerCreateInfo.minLod = 0.0f;
    samplerCreateInfo.maxLod = static_cast<float>(1);
    samplerCreateInfo.maxAnisotropy = 4.0;
    samplerCreateInfo.anisotropyEnable = VK_TRUE;
    samplerCreateInfo.borderColor = VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE;

    if (vkCreateSampler(m_device, &samplerCreateInfo, nullptr, &m_sampler) != VK_SUCCESS)
    {
        throw std::runtime_error("failed to create texture sampler!");
    }
};

void DefferedTexture::createForType(std::vector<VkImage>& images, std::vector<VkDeviceMemory>& memories, std::vector<VkImageView>& views, VkFormat format) {
    VkDeviceSize imageSize = WIDTH * HEIGHT * 4;
    if (format == VK_FORMAT_R16G16B16A16_SFLOAT) {
        imageSize *= sizeof(float);
    }
    
    for (int i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        images.resize(MAX_FRAMES_IN_FLIGHT);
        memories.resize(MAX_FRAMES_IN_FLIGHT);
        views.resize(MAX_FRAMES_IN_FLIGHT);

        VkBuffer stagingBuffer;
        VkDeviceMemory stagingBufferMemory;

        m_vulkanDevice.createBuffer(imageSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
            VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, stagingBuffer, stagingBufferMemory);

        // Images

        createImage(WIDTH, HEIGHT, format,
            VK_IMAGE_TILING_OPTIMAL,
            VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
            VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, images[i], memories[i]
        );

        transitionImageLayout(images[i], format, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);

        m_vulkanDevice.copyBufferToImage(stagingBuffer,
            images[i],
            static_cast<uint32_t>(WIDTH),
            static_cast<uint32_t>(HEIGHT),
            1
        );

        transitionImageLayout(images[i], format, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL);

        vkDestroyBuffer(m_device, stagingBuffer, nullptr);
        vkFreeMemory(m_device, stagingBufferMemory, nullptr);

        // Image views

        VkImageViewCreateInfo view{};
        view.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
        view.viewType = VK_IMAGE_VIEW_TYPE_2D;
        view.format = format;
        view.components = { VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_G, VK_COMPONENT_SWIZZLE_B, VK_COMPONENT_SWIZZLE_A };
        view.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        view.subresourceRange.baseMipLevel = 0;
        view.subresourceRange.baseArrayLayer = 0;
        view.subresourceRange.layerCount = 1;
        view.subresourceRange.levelCount = 1;
        view.image = images[i];

        if (vkCreateImageView(m_device, &view, nullptr, &views[i]) != VK_SUCCESS)
        {
            throw std::runtime_error("failed to create image view!");
        }
    }
}

void DefferedTexture::transitionImageLayout(VkImage image, VkFormat format, VkImageLayout oldLayout, VkImageLayout newLayout) {
    VkCommandBuffer commandBuffer = m_vulkanDevice.beginSingleTimeCommands();

    VkImageMemoryBarrier barrier{};
    barrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
    barrier.oldLayout = oldLayout;
    barrier.newLayout = newLayout;
    barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    barrier.image = image;
    barrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    barrier.subresourceRange.baseMipLevel = 0;
    barrier.subresourceRange.levelCount = 1;
    barrier.subresourceRange.baseArrayLayer = 0;
    barrier.subresourceRange.layerCount = 1;

    VkPipelineStageFlags sourceStage;
    VkPipelineStageFlags destinationStage;

    if (oldLayout == VK_IMAGE_LAYOUT_UNDEFINED && newLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL) {
        barrier.srcAccessMask = 0;
        barrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;

        sourceStage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
        destinationStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
    }
    else if (oldLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL && newLayout == VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL) {
        barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        barrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;

        sourceStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
        destinationStage = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT;
    }
    else {
        throw std::invalid_argument("unsupported layout transition!");
    }

    vkCmdPipelineBarrier(
        commandBuffer,
        sourceStage, destinationStage,
        0,
        0, nullptr,
        0, nullptr,
        1, &barrier
    );

    m_vulkanDevice.endSingleTimeCommands(commandBuffer);
}

void DefferedTexture::createImage(uint32_t width, uint32_t height, VkFormat format, VkImageTiling tiling, VkImageUsageFlags usage, VkMemoryPropertyFlags properties, VkImage& image, VkDeviceMemory& imageMemory) {
    VkImageCreateInfo imageInfo{};
    imageInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
    imageInfo.imageType = VK_IMAGE_TYPE_2D;
    imageInfo.extent.width = width;
    imageInfo.extent.height = height;
    imageInfo.extent.depth = 1;
    imageInfo.mipLevels = 1;
    imageInfo.arrayLayers = 1;
    imageInfo.format = format;
    imageInfo.tiling = tiling;
    imageInfo.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
    imageInfo.usage = usage;
    imageInfo.samples = VK_SAMPLE_COUNT_1_BIT;
    imageInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

    if (vkCreateImage(m_device, &imageInfo, nullptr, &image) != VK_SUCCESS) {
        throw std::runtime_error("failed to create image!");
    }

    VkMemoryRequirements memRequirements;
    vkGetImageMemoryRequirements(m_device, image, &memRequirements);

    VkMemoryAllocateInfo allocInfo{};
    allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    allocInfo.allocationSize = memRequirements.size;
    allocInfo.memoryTypeIndex = m_vulkanDevice.findMemoryType(memRequirements.memoryTypeBits, properties);

    if (vkAllocateMemory(m_device, &allocInfo, nullptr, &imageMemory) != VK_SUCCESS) {
        throw std::runtime_error("failed to allocate image memory!");
    }

    vkBindImageMemory(m_device, image, imageMemory, 0);
}

void DefferedTexture::cleanup() {
    if (!clean) {
        vkDestroySampler(m_device, m_sampler, nullptr);

        for (int i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
            vkDestroyImageView(m_device, m_albedoViews[i], nullptr);
            vkDestroyImage(m_device, m_albedoImages[i], nullptr);
            vkFreeMemory(m_device, m_albedoMemories[i], nullptr);

            vkDestroyImageView(m_device, m_normalViews[i], nullptr);
            vkDestroyImage(m_device, m_normalImages[i], nullptr);
            vkFreeMemory(m_device, m_normalMemories[i], nullptr);

            vkDestroyImageView(m_device, m_coefficientViews[i], nullptr);
            vkDestroyImage(m_device, m_coefficientImages[i], nullptr);
            vkFreeMemory(m_device, m_coefficientMemories[i], nullptr);

            vkDestroyImageView(m_device, m_materialViews[i], nullptr);
            vkDestroyImage(m_device, m_materialImages[i], nullptr);
            vkFreeMemory(m_device, m_materialMemories[i], nullptr);

            vkDestroyImageView(m_device, m_worldPosViews[i], nullptr);
            vkDestroyImage(m_device, m_worldPosImages[i], nullptr);
            vkFreeMemory(m_device, m_worldPosMemories[i], nullptr);
        }

        clean = true;
    }
}