#pragma once

#include "Common.h"
#include "VulkanDevice.h"
#include <stb_image.h>

/**
 *  A base class that encapsulates all related aspects for an objects material.
 */
class Material {
public:

    virtual void cleanup() {};

protected:
	Material(/*light parameters here? i.e. specular, metalic, etc.*/) {};

    
    bool clean = false;

  
};

/**
 *  A class representing an image texture in combination with a normal map as well as various parameters for usage in a pbr shader. 
 */
class ImageTexture : public Material {
public:
    ImageTexture() {};
    ImageTexture(VulkanDevice vulkanDevice, std::string texturePath, std::string normalMapPath, glm::vec3 coefficients /*x = ambient, y = diffuse, z = specular*/, float metallic, float roughness, float specularAlpha);
    ~ImageTexture() {};

    VkImageView& getTextureImageView();
    VkSampler& getTextureSampler();

    void loadNormalMap(std::string path);

    VkImageView& getNormalMapView() { return m_normalMapView; }
    VkSampler& getNormalMapSampler() { return m_normalMapSampler; }
    glm::vec3 getCoefficients() { return m_coefficients; }
    float getMetallic() { return m_metallic; }
    float getRoughness() { return m_roughness; }
    float getSpecularAlpha() { return m_specularAlpha; }

    void cleanup() override;

private:
    VulkanDevice m_vulkanDevice;
    VkDevice m_device;

    VkImage m_textureImage;
    VkDeviceMemory m_textureImageMemory;
    VkImageView m_textureImageView;
    VkSampler m_textureSampler;

    VkDeviceSize m_imageSize;
    int m_texWidth, m_texHeight, m_texChannels;
    stbi_uc* m_pixels;

    VkImage m_normalMap;
    VkDeviceMemory m_normalMapMemory;
    VkImageView m_normalMapView;
    VkSampler m_normalMapSampler;
    
    std::vector<glm::vec3> m_tangents;
    std::vector<glm::vec3> m_bitangents;

    glm::vec3 m_coefficients;
    float m_metallic;
    float m_roughness;
    float m_specularAlpha;

    void loadImage(std::string path);
    void createTextureImage();
    void createTextureImageView();
    void createTextureSampler(VkSampler* sampler);

    void createImage(uint32_t width, uint32_t height, VkFormat format,
        VkImageTiling tiling, VkImageUsageFlags usage, VkMemoryPropertyFlags properties, VkImage& image, VkDeviceMemory& imageMemory);
    VkImageView createImageView(VkImage image, VkFormat format, VkImageAspectFlags aspectFlags);
    void transitionImageLayout(VkImage image, VkFormat format, VkImageLayout oldLayout, VkImageLayout newLayout);
    

};

/**
 *  Enum representing all 6 sides of a cubemap.
 */
enum CubemapOrientations {
    POSX, NEGX, POSY, NEGY, POSZ, NEGZ
};

/**
 *  A base class representing cubemap textures.
 */
class CubemapTexture : public Material {
public:
    virtual VkImageView& getTextureImageView(int index) = 0; 
    virtual VkSampler& getTextureSampler() = 0;

    virtual void cleanup() override = 0;
    virtual ~CubemapTexture() = default;
protected:
    CubemapTexture() = default;

    int NUMBER_OF_CUBEMAP_IMAGES = 6;
};

/**
 *  A class representing a static cubemap texture.
 */
class StaticEnvironmentTexture : public CubemapTexture {
public:
    StaticEnvironmentTexture() = default;
    StaticEnvironmentTexture(VulkanDevice vulkanDevice, std::string path);
    ~StaticEnvironmentTexture() override = default;

    VkImageView& getTextureImageView(int index) override;
    VkSampler& getTextureSampler() override;

    void cleanup() override;

private:
    VulkanDevice m_vulkanDevice;
    VkDevice m_device;

    VkImage m_textureImage;
    VkDeviceMemory m_textureImageMemory;
    VkImageView m_textureImageView;
    VkSampler m_textureSampler;

    VkDeviceSize m_imageSize;
    int m_texWidth, m_texHeight, m_texChannels;
    stbi_uc* m_pixels[6];

    void loadImage(std::string path);
    void create();

    void createImage(uint32_t width, uint32_t height, VkFormat format,
        VkImageTiling tiling, VkImageUsageFlags usage, VkMemoryPropertyFlags properties,
        VkImage& image, VkDeviceMemory& imageMemory);
    void transitionImageLayout(VkImage image, VkFormat format, VkImageLayout oldLayout, VkImageLayout newLayout);
};

/**
 *  A class representing a dynamically updated cubemap texture.
 */
class DynEnvironmentTexture : public CubemapTexture {
public:
    DynEnvironmentTexture() = default;
    DynEnvironmentTexture(VulkanDevice vulkanDevice, uint32_t width, uint32_t height, uint32_t channels);
    ~DynEnvironmentTexture() override = default;

    VkImageView& getTextureImageView(int index) override;
    VkImageView& getDepthImageView(int index);
    VkSampler& getTextureSampler() override;
    std::vector<VkImageView> getAttachmentViews();
    std::vector<VkImageView> getDepthAttachmentViews();
    VkImage& getTextureImage(int index);
    uint32_t getWidth();
    uint32_t getHeight();

    void cleanup() override;

private:
    VulkanDevice m_vulkanDevice;
    VkDevice m_device;

    std::vector<VkImage> m_textureImages;
    std::vector<VkDeviceMemory> m_textureImageMemories;
    std::vector<VkImageView> m_cubemapImageViews;
    std::vector<VkImageView> m_colorAttachmentViews;

    std::vector<VkImage> m_depthImages;
    std::vector<VkDeviceMemory> m_depthImageMemories;
    std::vector<VkImageView> m_depthCubemapImageViews;
    std::vector<VkImageView> m_depthAttachmentViews;
    VkSampler m_textureSampler;

    VkDeviceSize m_imageSize;
    uint32_t m_texWidth, m_texHeight, m_texChannels;

    void create();

    void createImage(uint32_t width, uint32_t height, VkFormat format,
        VkImageTiling tiling, VkImageUsageFlags usage, VkMemoryPropertyFlags properties,
        VkImage& image, VkDeviceMemory& imageMemory);
    void transitionImageLayout(VkImage image, VkFormat format, VkImageLayout oldLayout, VkImageLayout newLayout, VkImageAspectFlagBits aspectFlags);
};

/**
 *  A class handling all g-buffer targets in a deffered rendering process.
 */
class DefferedTexture : public Material {
public:
    DefferedTexture() {};
    DefferedTexture(VulkanDevice vulkanDevice);
    ~DefferedTexture() {};

    VkImageView& getTextureImageView();
    VkSampler& getTextureSampler();

    void loadNormalMap(std::string path);

    std::vector<VkImageView> getAlbedoViews() { return m_albedoViews; }
    std::vector<VkImageView> getNormalViews() { return m_normalViews; }
    std::vector<VkImageView> getCoefficientViews() { return m_coefficientViews; }
    std::vector<VkImageView> getMaterialViews() { return m_materialViews; }
    std::vector<VkImageView> getWorldPosViews() { return m_worldPosViews; }
    std::vector<VkImage> getAlbedoImages() { return m_albedoImages; }
    std::vector<VkImage> getNormalImages() { return m_normalImages; }
    std::vector<VkImage> getCoefficientImages() { return m_coefficientImages; }
    std::vector<VkImage> getMaterialImages() { return m_materialImages; }
    std::vector<VkImage> getWorldPosImages() { return m_worldPosImages; }
    VkSampler getSampler() { return m_sampler; }

    void cleanup() override;

private:
    VulkanDevice m_vulkanDevice;
    VkDevice m_device;

    std::vector<VkImage> m_albedoImages;
    std::vector<VkImage> m_normalImages;
    std::vector<VkImage> m_coefficientImages;
    std::vector<VkImage> m_materialImages;
    std::vector<VkImage> m_worldPosImages;
    std::vector<VkDeviceMemory> m_albedoMemories;
    std::vector<VkDeviceMemory> m_normalMemories;
    std::vector<VkDeviceMemory> m_coefficientMemories;
    std::vector<VkDeviceMemory> m_materialMemories;
    std::vector<VkDeviceMemory> m_worldPosMemories;
    std::vector<VkImageView> m_albedoViews;
    std::vector<VkImageView> m_normalViews;
    std::vector<VkImageView> m_coefficientViews;
    std::vector<VkImageView> m_materialViews;
    std::vector<VkImageView> m_worldPosViews;
    VkSampler m_sampler;

    void create();
    void createForType(std::vector<VkImage>& images, std::vector<VkDeviceMemory>& memories, std::vector<VkImageView>& views, VkFormat format);

    void createImage(uint32_t width, uint32_t height, VkFormat format,
        VkImageTiling tiling, VkImageUsageFlags usage, VkMemoryPropertyFlags properties, VkImage& image, VkDeviceMemory& imageMemory);
    VkImageView createImageView(VkImage image, VkFormat format, VkImageAspectFlags aspectFlags);
    void transitionImageLayout(VkImage image, VkFormat format, VkImageLayout oldLayout, VkImageLayout newLayout);


};