#include "Object.h"

Object::Object(VulkanDevice vulkanDevice, Pipeline* pipeline, Model* model, glm::vec3 position, bool isMainObj) {
    m_vulkanDevice = vulkanDevice;
    m_device = m_vulkanDevice.getVkDevice();
    m_model = model;
    m_position = position;
    m_pipeline = pipeline;
	isMainObject = isMainObj;
}

VkBuffer& Object::getVertexBuffer() { return m_model->getVertexBuffer(); }
VkBuffer& Object::getIndexBuffer() { return m_model->getIndexBuffer(); }
VkDescriptorSet& Object::getDescriptorSet(int frame) { return m_descriptorSets.at(frame); }
std::vector<uint32_t> Object::getIndices() { return m_model->getIndices(); }

ImageObject::ImageObject(VulkanDevice vulkanDevice, Pipeline* pipeline, Model* model, ImageTexture* material, glm::vec3 position, float angle, glm::vec3 rotation, glm::vec3 scale, Path* path, float tilingFactor, bool isMainObj) {
    m_vulkanDevice = vulkanDevice;
    m_device = m_vulkanDevice.getVkDevice();
    m_model = model;
    m_position = position;
    m_pipeline = pipeline;
    m_material = material;
    m_scale = scale;
    m_rotation = rotation;
    m_angle = angle;
    m_tilingFactor = tilingFactor;
    isMainObject = isMainObj;

    if (!path) {
		m_hasPath = false;
	}
	else {
		m_hasPath = true;
		m_path = *path;
	}

    createUniformBuffers();
    createDescriptorPool();
    createDescriptorSets();
};

void ImageObject::update(float t, float dt, uint32_t currentImage, Camera& camera) {
    UniformBufferObject ubo{};
    glm::mat4 scaleMatrix = glm::scale(glm::mat4(1.0f), m_scale);
    glm::mat4 rotationMatrix = glm::rotate(glm::mat4(1.0f), m_angle, m_rotation);
    if (m_hasPath) {
        glm::mat4 totalRotation = m_additionalRotationMat * 
                                glm::rotate(glm::mat4(1.0f), m_angle, m_rotation);
        glm::quat rotation = m_path.getRotation(t, glm::vec3(0.f,0.f,-1.f));
        m_additionalRotationMat = glm::rotate(glm::mat4(1.f), glm::angle(rotation), glm::axis(rotation));
        glm::vec3 worldPos = m_path.getPosition(t);
        if (isMainObject) {
            camera.setLookAt(worldPos);
        }
        glm::mat4 worldToObject = glm::inverse(m_additionalRotationMat * rotationMatrix * scaleMatrix);
        m_position = glm::vec3(worldToObject * glm::vec4(worldPos, 1.0f));
    }
    glm::mat4 translationMatrix = glm::translate(glm::mat4(1.0f), m_position);
    glm::mat4 rotationScaleMatrix = m_additionalRotationMat * glm::rotate(glm::scale(glm::mat4(1.0f), m_scale), 
                                   m_angle, 
                                   m_rotation);

    
    ubo.model = rotationScaleMatrix * translationMatrix;
    ubo.view = camera.getViewMat();
    ubo.proj = camera.getProjMat();
    ubo.tilingFactor = m_tilingFactor;

    memcpy(m_uniformBuffersMapped[currentImage], &ubo, sizeof(ubo));

    MaterialUniformBufferObject mubo{};

    mubo.coefficients = m_material->getCoefficients();
    mubo.metallic = m_material->getMetallic();
    mubo.roughness = m_material->getRoughness();
    mubo.specularAlpha = m_material->getSpecularAlpha();

    memcpy(m_materialUniformBuffersMapped[currentImage], &mubo, sizeof(mubo));
}

void ImageObject::updateForCubemapDraw(float t, float dt, uint32_t currentImage, glm::mat4 viewMat, glm::mat4 projMat) {
    UniformBufferObject ubo{};
    glm::mat4 scaleMatrix = glm::scale(glm::mat4(1.0f), m_scale);
    glm::mat4 rotationMatrix = glm::rotate(glm::mat4(1.0f), m_angle, m_rotation);
    if (m_hasPath) {
        glm::mat4 totalRotation = m_additionalRotationMat *
            glm::rotate(glm::mat4(1.0f), m_angle, m_rotation);
        glm::quat rotation = m_path.getRotation(t, glm::vec3(0.f, 0.f, -1.f));
        m_additionalRotationMat = glm::rotate(glm::mat4(1.f), glm::angle(rotation), glm::axis(rotation));
        glm::vec3 worldPos = m_path.getPosition(t);
        glm::mat4 worldToObject = glm::inverse(m_additionalRotationMat * rotationMatrix * scaleMatrix);
        m_position = glm::vec3(worldToObject * glm::vec4(worldPos, 1.0f));
    }
    glm::mat4 translationMatrix = glm::translate(glm::mat4(1.0f), m_position);
    glm::mat4 rotationScaleMatrix = m_additionalRotationMat * glm::rotate(glm::scale(glm::mat4(1.0f), m_scale),
        m_angle,
        m_rotation);


    ubo.model = rotationScaleMatrix * translationMatrix;
    ubo.view = viewMat;
    ubo.proj = projMat;
    ubo.tilingFactor = m_tilingFactor;

    memcpy(m_uniformBuffersMapped[currentImage], &ubo, sizeof(ubo));

    MaterialUniformBufferObject mubo{};

    mubo.coefficients = m_material->getCoefficients();
    mubo.metallic = m_material->getMetallic();
    mubo.roughness = m_material->getRoughness();
    mubo.specularAlpha = m_material->getSpecularAlpha();

    memcpy(m_materialUniformBuffersMapped[currentImage], &mubo, sizeof(mubo));
};

void ImageObject::createUniformBuffers() {
    VkDeviceSize bufferSize = sizeof(UniformBufferObject);

    m_uniformBuffers.resize(MAX_FRAMES_IN_FLIGHT);
    m_uniformBuffersMemory.resize(MAX_FRAMES_IN_FLIGHT);
    m_uniformBuffersMapped.resize(MAX_FRAMES_IN_FLIGHT);

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        m_vulkanDevice.createBuffer(bufferSize, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, m_uniformBuffers[i], m_uniformBuffersMemory[i]);

        vkMapMemory(m_device, m_uniformBuffersMemory[i], 0, bufferSize, 0, &m_uniformBuffersMapped[i]);
    }

    VkDeviceSize materialBufferSize = sizeof(MaterialUniformBufferObject);

    m_materialUniformBuffers.resize(MAX_FRAMES_IN_FLIGHT);
    m_materialUniformBuffersMemory.resize(MAX_FRAMES_IN_FLIGHT);
    m_materialUniformBuffersMapped.resize(MAX_FRAMES_IN_FLIGHT);

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        m_vulkanDevice.createBuffer(materialBufferSize, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, m_materialUniformBuffers[i], m_materialUniformBuffersMemory[i]);

        vkMapMemory(m_device, m_materialUniformBuffersMemory[i], 0, materialBufferSize, 0, &m_materialUniformBuffersMapped[i]);
    }
}

void ImageObject::createDescriptorPool() {
    std::array<VkDescriptorPoolSize, 4> poolSizes{};
    poolSizes[0].type = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
    poolSizes[0].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);
    poolSizes[1].type = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
    poolSizes[1].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);
	poolSizes[2].type = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
	poolSizes[2].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);
    poolSizes[3].type = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
    poolSizes[3].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    VkDescriptorPoolCreateInfo poolInfo{};
    poolInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
    poolInfo.poolSizeCount = static_cast<uint32_t>(poolSizes.size());
    poolInfo.pPoolSizes = poolSizes.data();
    poolInfo.maxSets = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    if (vkCreateDescriptorPool(m_device, &poolInfo, nullptr, &m_descriptorPool) != VK_SUCCESS) {
        throw std::runtime_error("failed to create descriptor pool!");
    }
}

void ImageObject::createDescriptorSets() {
    std::vector<VkDescriptorSetLayout> layouts(MAX_FRAMES_IN_FLIGHT, m_pipeline->getDescriptorSetLayout());
    VkDescriptorSetAllocateInfo allocInfo{};
    allocInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    allocInfo.descriptorPool = m_descriptorPool;
    allocInfo.descriptorSetCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);
    allocInfo.pSetLayouts = layouts.data();

    m_descriptorSets.resize(MAX_FRAMES_IN_FLIGHT);
    if (vkAllocateDescriptorSets(m_device, &allocInfo, m_descriptorSets.data()) != VK_SUCCESS) {
        throw std::runtime_error("failed to allocate descriptor sets!");
    }

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        VkDescriptorBufferInfo bufferInfo{};
        bufferInfo.buffer = m_uniformBuffers[i];
        bufferInfo.offset = 0;
        bufferInfo.range = sizeof(UniformBufferObject);

        VkDescriptorImageInfo imageInfo{};
        imageInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        imageInfo.imageView = m_material->getTextureImageView();
        imageInfo.sampler = m_material->getTextureSampler();

        VkDescriptorImageInfo normalMapInfo{};
        normalMapInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        normalMapInfo.imageView = m_material->getNormalMapView();
        normalMapInfo.sampler = m_material->getNormalMapSampler();

        VkDescriptorBufferInfo materialBufferInfo{};
        materialBufferInfo.buffer = m_materialUniformBuffers[i];
        materialBufferInfo.offset = 0;
        materialBufferInfo.range = sizeof(MaterialUniformBufferObject);

        std::array<VkWriteDescriptorSet, 4> descriptorWrites{};

        descriptorWrites[0].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[0].dstSet = m_descriptorSets[i];
        descriptorWrites[0].dstBinding = 0;
        descriptorWrites[0].dstArrayElement = 0;
        descriptorWrites[0].descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
        descriptorWrites[0].descriptorCount = 1;
        descriptorWrites[0].pBufferInfo = &bufferInfo;

        descriptorWrites[1].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[1].dstSet = m_descriptorSets[i];
        descriptorWrites[1].dstBinding = 1;
        descriptorWrites[1].dstArrayElement = 0;
        descriptorWrites[1].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        descriptorWrites[1].descriptorCount = 1;
        descriptorWrites[1].pImageInfo = &imageInfo;

        descriptorWrites[2].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[2].dstSet = m_descriptorSets[i];
        descriptorWrites[2].dstBinding = 2;
        descriptorWrites[2].dstArrayElement = 0;
        descriptorWrites[2].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        descriptorWrites[2].descriptorCount = 1;
        descriptorWrites[2].pImageInfo = &normalMapInfo;

        descriptorWrites[3].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[3].dstSet = m_descriptorSets[i];
        descriptorWrites[3].dstBinding = 3;
        descriptorWrites[3].dstArrayElement = 0;
        descriptorWrites[3].descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
        descriptorWrites[3].descriptorCount = 1;
        descriptorWrites[3].pBufferInfo = &materialBufferInfo;

        vkUpdateDescriptorSets(m_device, static_cast<uint32_t>(descriptorWrites.size()), descriptorWrites.data(), 0, nullptr);
    }
}

void ImageObject::destroy() {
    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        vkDestroyBuffer(m_device, m_uniformBuffers[i], nullptr);
        vkFreeMemory(m_device, m_uniformBuffersMemory[i], nullptr);
    }
    vkDestroyDescriptorPool(m_device, m_descriptorPool, nullptr);   
}

void ImageObject::cleanup() {
    m_material->cleanup();
    m_model->cleanup();
    
    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        vkDestroyBuffer(m_device, m_uniformBuffers[i], nullptr);
        vkFreeMemory(m_device, m_uniformBuffersMemory[i], nullptr);
        vkDestroyBuffer(m_device, m_materialUniformBuffers[i], nullptr);
        vkFreeMemory(m_device, m_materialUniformBuffersMemory[i], nullptr);
    }
    vkDestroyDescriptorPool(m_device, m_descriptorPool, nullptr);
}

ParticleObject::ParticleObject(VulkanDevice vulkanDevice, Pipeline* pipeline, Model* model, glm::vec3 position, ParticleSystem* particleSystem, Pipeline* drawingPipeline, bool isMainObj) {
    m_vulkanDevice = vulkanDevice;
    m_device = vulkanDevice.getVkDevice();
    m_pipeline = pipeline;
    m_model = model;
    m_position = position;
    m_particleSystem = particleSystem;
    m_drawingPipeline = drawingPipeline;
    isMainObject = isMainObj;

    createUniformBuffers();
    createDescriptorPool();
    createDescriptorSets();
    createDrawingUniformBuffers();
    createDrawingDescriptorPool();
    createDrawingDescriptorSets();
}

void ParticleObject::createDescriptorSets() {
    std::vector<VkDescriptorSetLayout> layouts(MAX_FRAMES_IN_FLIGHT, m_pipeline->getDescriptorSetLayout());
    VkDescriptorSetAllocateInfo allocInfo{};
    allocInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    allocInfo.descriptorPool = m_descriptorPool;
    allocInfo.descriptorSetCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);
    allocInfo.pSetLayouts = layouts.data();

    m_descriptorSets.resize(MAX_FRAMES_IN_FLIGHT);
    if (vkAllocateDescriptorSets(m_device, &allocInfo, m_descriptorSets.data()) != VK_SUCCESS) {
        throw std::runtime_error("failed to allocate descriptor sets!");
    }

    std::vector<VkBuffer> storageBuffers = m_particleSystem->getStorageBuffers();

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        VkDescriptorBufferInfo uniformBufferInfo{};
        uniformBufferInfo.buffer = m_uniformBuffers[i];
        uniformBufferInfo.offset = 0;
        uniformBufferInfo.range = sizeof(ParticleUniformBufferObject);

        std::array<VkWriteDescriptorSet, 3> descriptorWrites{};
        descriptorWrites[0].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[0].dstSet = m_descriptorSets[i];
        descriptorWrites[0].dstBinding = 0;
        descriptorWrites[0].dstArrayElement = 0;
        descriptorWrites[0].descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
        descriptorWrites[0].descriptorCount = 1;
        descriptorWrites[0].pBufferInfo = &uniformBufferInfo;

        VkDescriptorBufferInfo storageBufferInfoLastFrame{};
        storageBufferInfoLastFrame.buffer = m_particleSystem->getStorageBuffers()[(i - 1) % MAX_FRAMES_IN_FLIGHT];
        storageBufferInfoLastFrame.offset = 0;
        storageBufferInfoLastFrame.range = sizeof(Particle) * m_particleSystem->getParticleCount();

        descriptorWrites[1].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[1].dstSet = m_descriptorSets[i];
        descriptorWrites[1].dstBinding = 1;
        descriptorWrites[1].dstArrayElement = 0;
        descriptorWrites[1].descriptorType = VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
        descriptorWrites[1].descriptorCount = 1;
        descriptorWrites[1].pBufferInfo = &storageBufferInfoLastFrame;

        VkDescriptorBufferInfo storageBufferInfoCurrentFrame{};
        storageBufferInfoCurrentFrame.buffer = m_particleSystem->getStorageBuffers()[i];
        storageBufferInfoCurrentFrame.offset = 0;
        storageBufferInfoCurrentFrame.range = sizeof(Particle) * m_particleSystem->getParticleCount();

        descriptorWrites[2].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[2].dstSet = m_descriptorSets[i];
        descriptorWrites[2].dstBinding = 2;
        descriptorWrites[2].dstArrayElement = 0;
        descriptorWrites[2].descriptorType = VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
        descriptorWrites[2].descriptorCount = 1;
        descriptorWrites[2].pBufferInfo = &storageBufferInfoCurrentFrame;

        vkUpdateDescriptorSets(m_device, 3, descriptorWrites.data(), 0, nullptr);
    }
}

void ParticleObject::createDescriptorPool() {
    std::array<VkDescriptorPoolSize, 2> poolSizes{};
    poolSizes[0].type = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
    poolSizes[0].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    poolSizes[1].type = VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
    poolSizes[1].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT) * 2;

    VkDescriptorPoolCreateInfo poolInfo{};
    poolInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
    poolInfo.poolSizeCount = 2;
    poolInfo.pPoolSizes = poolSizes.data();
    poolInfo.maxSets = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    if (vkCreateDescriptorPool(m_device, &poolInfo, nullptr, &m_descriptorPool) != VK_SUCCESS) {
        throw std::runtime_error("failed to create descriptor pool!");
    }
}

void ParticleObject::createUniformBuffers() {
    VkDeviceSize bufferSize = sizeof(ParticleUniformBufferObject);

    m_uniformBuffers.resize(MAX_FRAMES_IN_FLIGHT);
    m_uniformBuffersMemory.resize(MAX_FRAMES_IN_FLIGHT);
    m_uniformBuffersMapped.resize(MAX_FRAMES_IN_FLIGHT);

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        m_vulkanDevice.createBuffer(bufferSize, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, m_uniformBuffers[i], m_uniformBuffersMemory[i]);

        vkMapMemory(m_device, m_uniformBuffersMemory[i], 0, bufferSize, 0, &m_uniformBuffersMapped[i]);
    }
}

void ParticleObject::update(float t, float dt, uint32_t currentImage, Camera& camera) {
    ParticleUniformBufferObject pubo{};
    pubo.model = glm::translate(glm::mat4(1.0f), glm::vec3(20.0f, 0.0f, 0.0f));
    pubo.view = camera.getViewMat();
    pubo.proj = camera.getProjMat();
    pubo.deltaTime = dt;

    memcpy(m_uniformBuffersMapped[currentImage], &pubo, sizeof(pubo));

    UniformBufferObject ubo{};
    ubo.model = glm::mat4(1.0f);
    ubo.view = camera.getViewMat();
    ubo.proj = camera.getProjMat();

    memcpy(m_drawingUniformBuffersMapped[currentImage], &ubo, sizeof(ubo));
};

void ParticleObject::updateForCubemapDraw(float t, float dt, uint32_t currentImage, glm::mat4 viewMat, glm::mat4 projMat) {
    ParticleUniformBufferObject pubo{};
    pubo.model = glm::translate(glm::mat4(1.0f), glm::vec3(20.0f, 0.0f, 0.0f));
    pubo.view = viewMat;
    pubo.proj = projMat;
    pubo.deltaTime = dt;

    memcpy(m_uniformBuffersMapped[currentImage], &pubo, sizeof(pubo));

    UniformBufferObject ubo{};
    ubo.model = glm::mat4(1.0f);
    ubo.view = viewMat;
    ubo.proj = projMat;

    memcpy(m_drawingUniformBuffersMapped[currentImage], &ubo, sizeof(ubo));
};

void ParticleObject::destroy() {
    // nothing yet
};

void ParticleObject::cleanup() {
    m_particleSystem->cleanup();
    m_model->cleanup();

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        vkDestroyBuffer(m_device, m_uniformBuffers[i], nullptr);
        vkFreeMemory(m_device, m_uniformBuffersMemory[i], nullptr);
        vkDestroyBuffer(m_device, m_drawingUniformBuffers[i], nullptr);
        vkFreeMemory(m_device, m_drawingUniformBuffersMemory[i], nullptr);
    }
    vkDestroyDescriptorPool(m_device, m_descriptorPool, nullptr);
    vkDestroyDescriptorPool(m_device, m_drawingDescriptorPool, nullptr);
};

void ParticleObject::createDrawingUniformBuffers() {
    VkDeviceSize bufferSize = sizeof(UniformBufferObject);

    m_drawingUniformBuffers.resize(MAX_FRAMES_IN_FLIGHT);
    m_drawingUniformBuffersMemory.resize(MAX_FRAMES_IN_FLIGHT);
    m_drawingUniformBuffersMapped.resize(MAX_FRAMES_IN_FLIGHT);

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        m_vulkanDevice.createBuffer(bufferSize, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, m_drawingUniformBuffers[i], m_drawingUniformBuffersMemory[i]);

        vkMapMemory(m_device, m_drawingUniformBuffersMemory[i], 0, bufferSize, 0, &m_drawingUniformBuffersMapped[i]);
    }
};

void ParticleObject::createDrawingDescriptorPool() {
    std::array<VkDescriptorPoolSize, 1> poolSizes{};
    poolSizes[0].type = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
    poolSizes[0].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    VkDescriptorPoolCreateInfo poolInfo{};
    poolInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
    poolInfo.poolSizeCount = static_cast<uint32_t>(poolSizes.size());
    poolInfo.pPoolSizes = poolSizes.data();
    poolInfo.maxSets = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    if (vkCreateDescriptorPool(m_device, &poolInfo, nullptr, &m_drawingDescriptorPool) != VK_SUCCESS) {
        throw std::runtime_error("failed to create descriptor pool!");
    }
};

void ParticleObject::createDrawingDescriptorSets() {
    std::vector<VkDescriptorSetLayout> layouts(MAX_FRAMES_IN_FLIGHT, m_drawingPipeline->getDescriptorSetLayout());
    VkDescriptorSetAllocateInfo allocInfo{};
    allocInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    allocInfo.descriptorPool = m_drawingDescriptorPool;
    allocInfo.descriptorSetCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);
    allocInfo.pSetLayouts = layouts.data();

    m_drawingDescriptorSets.resize(MAX_FRAMES_IN_FLIGHT);
    if (vkAllocateDescriptorSets(m_device, &allocInfo, m_drawingDescriptorSets.data()) != VK_SUCCESS) {
        throw std::runtime_error("failed to allocate descriptor sets!");
    }

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        VkDescriptorBufferInfo bufferInfo{};
        bufferInfo.buffer = m_drawingUniformBuffers[i];
        bufferInfo.offset = 0;
        bufferInfo.range = sizeof(UniformBufferObject);

        std::array<VkWriteDescriptorSet, 1> descriptorWrites{};

        descriptorWrites[0].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[0].dstSet = m_drawingDescriptorSets[i];
        descriptorWrites[0].dstBinding = 0;
        descriptorWrites[0].dstArrayElement = 0;
        descriptorWrites[0].descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
        descriptorWrites[0].descriptorCount = 1;
        descriptorWrites[0].pBufferInfo = &bufferInfo;

        vkUpdateDescriptorSets(m_device, static_cast<uint32_t>(descriptorWrites.size()), descriptorWrites.data(), 0, nullptr);
    }
};

DefferedDrawObject::DefferedDrawObject(VulkanDevice vulkanDevice, Pipeline* pipeline, Model* model, glm::vec3 position, DefferedTexture* material, bool isMainObj) {
    m_vulkanDevice = vulkanDevice;
    m_device = vulkanDevice.getVkDevice();
    m_pipeline = pipeline;
    m_model = model;
    m_position = position;
    m_material = material;
    isMainObject = isMainObj;

    createUniformBuffers();
    createDescriptorPool();
    createDescriptorSets();
};

void DefferedDrawObject::createUniformBuffers() {
    VkDeviceSize bufferSize = sizeof(LightUniformBufferObject);

    m_uniformBuffers.resize(MAX_FRAMES_IN_FLIGHT);
    m_uniformBuffersMemory.resize(MAX_FRAMES_IN_FLIGHT);
    m_uniformBuffersMapped.resize(MAX_FRAMES_IN_FLIGHT);

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        m_vulkanDevice.createBuffer(bufferSize, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, m_uniformBuffers[i], m_uniformBuffersMemory[i]);

        vkMapMemory(m_device, m_uniformBuffersMemory[i], 0, bufferSize, 0, &m_uniformBuffersMapped[i]);
    }
}

void DefferedDrawObject::createDescriptorPool() {
    std::array<VkDescriptorPoolSize, 6> poolSizes{};
    poolSizes[0].type = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
    poolSizes[0].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    poolSizes[1].type = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
    poolSizes[1].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    poolSizes[2].type = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
    poolSizes[2].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    poolSizes[3].type = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
    poolSizes[3].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    poolSizes[4].type = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
    poolSizes[4].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    poolSizes[5].type = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
    poolSizes[5].descriptorCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    VkDescriptorPoolCreateInfo poolInfo{};
    poolInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
    poolInfo.poolSizeCount = static_cast<uint32_t>(poolSizes.size());
    poolInfo.pPoolSizes = poolSizes.data();
    poolInfo.maxSets = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);

    if (vkCreateDescriptorPool(m_device, &poolInfo, nullptr, &m_descriptorPool) != VK_SUCCESS) {
        throw std::runtime_error("failed to create descriptor pool!");
    }
}

void DefferedDrawObject::createDescriptorSets() {
    std::vector<VkDescriptorSetLayout> layouts(MAX_FRAMES_IN_FLIGHT, m_pipeline->getDescriptorSetLayout());
    VkDescriptorSetAllocateInfo allocInfo{};
    allocInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    allocInfo.descriptorPool = m_descriptorPool;
    allocInfo.descriptorSetCount = static_cast<uint32_t>(MAX_FRAMES_IN_FLIGHT);
    allocInfo.pSetLayouts = layouts.data();

    m_descriptorSets.resize(MAX_FRAMES_IN_FLIGHT);
    if (vkAllocateDescriptorSets(m_device, &allocInfo, m_descriptorSets.data()) != VK_SUCCESS) {
        throw std::runtime_error("failed to allocate descriptor sets!");
    }

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        VkDescriptorBufferInfo lightBufferInfo{};
        lightBufferInfo.buffer = m_uniformBuffers[i];
        lightBufferInfo.offset = 0;
        lightBufferInfo.range = sizeof(LightUniformBufferObject);

        VkDescriptorImageInfo albedoInfo{};
        albedoInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        albedoInfo.imageView = m_material->getAlbedoViews()[i];
        albedoInfo.sampler = m_material->getSampler();

        VkDescriptorImageInfo normalInfo{};
        normalInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        normalInfo.imageView = m_material->getNormalViews()[i];
        normalInfo.sampler = m_material->getSampler();

        VkDescriptorImageInfo coefficientsInfo{};
        coefficientsInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        coefficientsInfo.imageView = m_material->getCoefficientViews()[i];
        coefficientsInfo.sampler = m_material->getSampler();

        VkDescriptorImageInfo materialInfo{};
        materialInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        materialInfo.imageView = m_material->getMaterialViews()[i];
        materialInfo.sampler = m_material->getSampler();

        VkDescriptorImageInfo worldPosInfo{};
        worldPosInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        worldPosInfo.imageView = m_material->getWorldPosViews()[i];
        worldPosInfo.sampler = m_material->getSampler();

        std::array<VkWriteDescriptorSet, 6> descriptorWrites{};

        descriptorWrites[0].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[0].dstSet = m_descriptorSets[i];
        descriptorWrites[0].dstBinding = 0;
        descriptorWrites[0].dstArrayElement = 0;
        descriptorWrites[0].descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
        descriptorWrites[0].descriptorCount = 1;
        descriptorWrites[0].pBufferInfo = &lightBufferInfo;

        descriptorWrites[1].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[1].dstSet = m_descriptorSets[i];
        descriptorWrites[1].dstBinding = 1;
        descriptorWrites[1].dstArrayElement = 0;
        descriptorWrites[1].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        descriptorWrites[1].descriptorCount = 1;
        descriptorWrites[1].pImageInfo = &albedoInfo;

        descriptorWrites[2].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[2].dstSet = m_descriptorSets[i];
        descriptorWrites[2].dstBinding = 2;
        descriptorWrites[2].dstArrayElement = 0;
        descriptorWrites[2].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        descriptorWrites[2].descriptorCount = 1;
        descriptorWrites[2].pImageInfo = &normalInfo;

        descriptorWrites[3].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[3].dstSet = m_descriptorSets[i];
        descriptorWrites[3].dstBinding = 3;
        descriptorWrites[3].dstArrayElement = 0;
        descriptorWrites[3].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        descriptorWrites[3].descriptorCount = 1;
        descriptorWrites[3].pImageInfo = &coefficientsInfo;

        descriptorWrites[4].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[4].dstSet = m_descriptorSets[i];
        descriptorWrites[4].dstBinding = 4;
        descriptorWrites[4].dstArrayElement = 0;
        descriptorWrites[4].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        descriptorWrites[4].descriptorCount = 1;
        descriptorWrites[4].pImageInfo = &materialInfo;

        descriptorWrites[5].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        descriptorWrites[5].dstSet = m_descriptorSets[i];
        descriptorWrites[5].dstBinding = 5;
        descriptorWrites[5].dstArrayElement = 0;
        descriptorWrites[5].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        descriptorWrites[5].descriptorCount = 1;
        descriptorWrites[5].pImageInfo = &worldPosInfo;

        vkUpdateDescriptorSets(m_device, static_cast<uint32_t>(descriptorWrites.size()), descriptorWrites.data(), 0, nullptr);
    }
}

void DefferedDrawObject::update(float t, float dt, uint32_t currentImage, Camera& camera, DirectionalLight& light) {
    LightUniformBufferObject ubo{};
    ubo.cameraWorldPos = camera.getPosition();
    ubo.dirLColor = light.getColor();
    ubo.dirLDirection = light.getDirection();

    memcpy(m_uniformBuffersMapped[currentImage], &ubo, sizeof(ubo));
}

void DefferedDrawObject::update(float t, float dt, uint32_t currentImage, Camera& camera) {
    //nothing
}

void DefferedDrawObject::updateForCubemapDraw(float t, float dt, uint32_t currentImage, glm::mat4 viewMat, glm::mat4 projMat) {
    //nothing
}

void DefferedDrawObject::cleanup() {
    m_material->cleanup();
    m_model->cleanup();

    for (size_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
        vkDestroyBuffer(m_device, m_uniformBuffers[i], nullptr);
        vkFreeMemory(m_device, m_uniformBuffersMemory[i], nullptr);
    }
    vkDestroyDescriptorPool(m_device, m_descriptorPool, nullptr);
}

void DefferedDrawObject::destroy() {
    //nothing
}