#pragma once

#include "Common.h"
#include "Material.h"
#include "Model.h"
#include "Pipeline.h"
#include "Camera.h"
#include "ParticleSystem.h"
#include "Path.h"
#include "Light.h"

/**
 *  A base class that encapsulates all necessary aspects for drawing the object.
 */
class Object {
public:
	void setAdditionalRotationMat(glm::mat4 rot) { m_additionalRotationMat = rot; }

	VkBuffer& getVertexBuffer();
	VkBuffer& getIndexBuffer();
	VkDescriptorSet& getDescriptorSet(int frame);
	std::vector<uint32_t> getIndices();
	glm::vec3 getPosition() { return m_position; }

	virtual void destroy() {};
	virtual void update(float t, float dt, uint32_t currentImage, Camera& camera) {};
	virtual void updateForCubemapDraw(float t, float dt, uint32_t currentImage, glm::mat4 viewMat, glm::mat4 projMat) {};
	virtual void cleanup() {};

protected:
	Model* m_model;
	VulkanDevice m_vulkanDevice;
	VkDevice m_device;
	Pipeline* m_pipeline;

	glm::vec3 m_position;
	glm::vec3 m_vdirection;
	glm::vec3 m_scale;
	glm::vec3 m_rotation;
	float m_angle;
	glm::mat4 m_model_mat = glm::mat4(1.0f);
	glm::mat4 m_additionalRotationMat = glm::mat4(1.0f);

	std::vector<VkBuffer> m_uniformBuffers;
	std::vector<VkDeviceMemory> m_uniformBuffersMemory;
	std::vector<void*> m_uniformBuffersMapped;

	VkDescriptorPool m_descriptorPool;
	std::vector<VkDescriptorSet> m_descriptorSets;

	Path m_path;
	bool m_hasPath = false;
	glm::vec3 m_forward = glm::vec3(0.0f, 0.0f, -1.0f);

	bool isMainObject;

	Object() {};
	Object(VulkanDevice device, Pipeline* pipeline, Model* model, glm::vec3 position, bool isMainObj = false);
};

/**
 *  A class representing a geometry with an image texture and normal map.
 */
class ImageObject : public Object {
public:
	ImageObject() {};
	ImageObject(VulkanDevice vulkanDevice, Pipeline* pipeline, Model* model, ImageTexture* material, glm::vec3 position, float angle, glm::vec3 rotation, glm::vec3 scale, Path* path, float tilingFactor, bool isMainObj = false);
	~ImageObject() {};

	void update(float t, float dt, uint32_t currentImage, Camera& camera) override;
	void updateForCubemapDraw(float t, float dt, uint32_t currentImage, glm::mat4 viewMat, glm::mat4 projMat) override;
	void destroy() override;
	void cleanup() override;

private:

	ImageTexture* m_material;
	float m_tilingFactor;

	std::vector<VkBuffer> m_materialUniformBuffers;
	std::vector<VkDeviceMemory> m_materialUniformBuffersMemory;
	std::vector<void*> m_materialUniformBuffersMapped;

	void createUniformBuffers();
	void createDescriptorPool();
	void createDescriptorSets();
};


/**
 *  A class representing a GPU particle system.
 */
class ParticleObject : public Object {
public:
	ParticleObject() {};
	ParticleObject(VulkanDevice vulkanDevice, Pipeline* pipeline, Model* model, glm::vec3 position, ParticleSystem* particleSystem, Pipeline* drawingPipeline, bool isMainObj = false);

	void update(float t, float dt, uint32_t currentImage, Camera& camera) override;
	void updateForCubemapDraw(float t, float dt, uint32_t currentImage, glm::mat4 viewMat, glm::mat4 projMat) override;
	void destroy() override;
	void cleanup() override;

	Pipeline& getComputePipeline() const { return *m_pipeline; }
	Pipeline& getDrawingPipeline() const { return *m_drawingPipeline; }
	VkPipeline getComputeVkPipeline() const { return m_pipeline->getPipeline(); }
	VkPipeline getDrawingVkPipeline() const { return m_drawingPipeline->getPipeline(); }
	VkPipeline getDrawToCubeVkPipeline() const { return m_drawToCubePipeline->getPipeline(); }
	VkPipelineLayout getComputePipelineLayout() const { return m_pipeline->getPipelineLayout(); }
	VkPipelineLayout getDrawingPipelineLayout() const { return m_drawingPipeline->getPipelineLayout(); }
	VkPipelineLayout getDrawToCubePipelineLayout() const { return m_drawToCubePipeline->getPipelineLayout(); }
	int getParticleCount() const { return m_particleSystem->getParticleCount(); }
	std::vector<VkBuffer> getStorageBuffers() const { return m_particleSystem->getStorageBuffers(); }
	void setDrawToCubePipeline(Pipeline* drawToCubePipeline) { m_drawToCubePipeline = drawToCubePipeline; }

	std::vector<VkDescriptorSet> getDrawingDescriptorSets() const { return m_drawingDescriptorSets; }

private:
	ParticleSystem* m_particleSystem;
	Pipeline* m_drawingPipeline;
	Pipeline* m_drawToCubePipeline;

	std::vector<VkBuffer> m_drawingUniformBuffers;
	std::vector<VkDeviceMemory> m_drawingUniformBuffersMemory;
	std::vector<void*> m_drawingUniformBuffersMapped;

	VkDescriptorPool m_drawingDescriptorPool;
	std::vector<VkDescriptorSet> m_drawingDescriptorSets;

	void createUniformBuffers();
	void createDescriptorPool();
	void createDescriptorSets();
	void createDrawingUniformBuffers();
	void createDrawingDescriptorPool();
	void createDrawingDescriptorSets();
};

/**
 *  A class representing the full screen draw object in a deffered rendering process.
 */
class DefferedDrawObject : public Object {
public:
	DefferedDrawObject() {};
	DefferedDrawObject(VulkanDevice vulkanDevice, Pipeline* pipeline, Model* model, glm::vec3 position, DefferedTexture* material, bool isMainObj = false);
	~DefferedDrawObject() {};

	void update(float t, float dt, uint32_t currentImage, Camera& camera) override;
	void update(float t, float dt, uint32_t currentImage, Camera& camera, DirectionalLight& light);
	void updateForCubemapDraw(float t, float dt, uint32_t currentImage, glm::mat4 viewMat, glm::mat4 projMat) override;
	void destroy() override;
	void cleanup() override;

	DefferedTexture* getMaterial() { return m_material; }
	VkPipeline getPipeline() { return m_pipeline->getPipeline(); }
	VkPipelineLayout getPipelineLayout() { return m_pipeline->getPipelineLayout(); }
	VkDescriptorSet& getDescriptorSet(int currentFrame) { return m_descriptorSets[currentFrame]; }

private:

	DefferedTexture* m_material;

	void createUniformBuffers();
	void createDescriptorPool();
	void createDescriptorSets();
};

