#pragma once

#include "Common.h"
#include "VulkanDevice.h"

/**
 *  A base class that represents a Vulkan pipeline.
 */
class Pipeline {
public:
	~Pipeline() {};

	VkPipeline getPipeline();
	VkPipelineLayout getPipelineLayout();
	VkDescriptorSetLayout getDescriptorSetLayout();

	void cleanup();

protected:
	Pipeline() {};
	Pipeline(std::string vertShader_path, std::string fragShader_path, VulkanDevice vulkanDevice, VkRenderPass renderpass);	

	VulkanDevice m_vulkanDevice;
	VkDevice m_device;
	VkRenderPass m_renderPass;
	
	VkPipeline m_pipeline;
	VkPipelineLayout m_pipelineLayout;
    VkDescriptorSetLayout m_descriptorSetLayout;

	VkShaderModule createShaderModule(const std::vector<char>& code);
	std::vector<char> readFile(const std::string& filename);
	
	virtual void createDescriptorSetLayout() {};
};

/**
 *  A class representing the render pipeline of ImageObjects.
 */
class ImagePipeline : public Pipeline {
public:
	ImagePipeline() {};
	ImagePipeline(std::string vertShader_path, std::string fragShader_path, VulkanDevice vulkanDevice, VkRenderPass renderpass, int attachmentCount);

private:
	void createDescriptorSetLayout() override;

};

/**
 *  A class representing the render pipeline of Cubemaps.
 */
class CubemapPipeline : public Pipeline {
public:
	CubemapPipeline() {};
	CubemapPipeline(std::string vertShader_path, std::string fragShader_path, VulkanDevice vulkanDevice, VkRenderPass renderpass, int attachmentCount);

private:
	void createDescriptorSetLayout() override;

};

/**
 *  A class representing a compute pipeline.
 */
class ComputePipeline : public Pipeline {
public:
	ComputePipeline() {};
	ComputePipeline(std::string compShader_path, VulkanDevice vulkanDevice, VkRenderPass renderpass);

private:
	void createDescriptorSetLayout() override;
};

/**
 *  A class representing the render pipeline of a GPU ParticleSystem.
 */
class ParticleDrawPipeline : public Pipeline {
public:
	ParticleDrawPipeline() {};
	ParticleDrawPipeline(std::string vertShader_path, std::string fragShader_path, std::string geomShader_path, VulkanDevice vulkanDevice, VkRenderPass renderpass, int attachmentCount);

private:
	void createDescriptorSetLayout() override;
};

/**
 *  A class representing the render pipeline of Cubemaps with dynamic texture.
 */
class DynEnvironmentMapDrawPipeline : public Pipeline {
public:
	DynEnvironmentMapDrawPipeline() {};
	DynEnvironmentMapDrawPipeline(std::string vertShader_path, std::string fragShader_path, VulkanDevice vulkanDevice, VkRenderPass renderpass);

private:
	void createDescriptorSetLayout() override;
};

/**
 *  A class representing the render pipeline for rendering into Cubemap Textures.
 */
class RenderToCubemapPipeline : public Pipeline {
public:
	RenderToCubemapPipeline() {};
	RenderToCubemapPipeline(std::string vertShader_path, std::string fragShader_path, VulkanDevice vulkanDevice, VkRenderPass renderpass);

private:
	void createDescriptorSetLayout() override;
};

/**
 *  A class representing the render pipeline of the final draw in a deffered rendering process.
 */
class RenderDefferedPipeline : public Pipeline {
public:
	RenderDefferedPipeline() {};
	RenderDefferedPipeline(std::string vertShader_path, std::string fragShader_path, VulkanDevice vulkanDevice, VkRenderPass renderpass);

private:
	void createDescriptorSetLayout() override;
};