#pragma once

#define GLFW_INCLUDE_VULKAN
#include <glfw3.h>

#define GLM_FORCE_RADIANS
#define GLM_FORCE_DEPTH_ZERO_TO_ONE
#define GLM_ENABLE_EXPERIMENTAL


#include "Common.h"
#include "ObjectHandler.h"
#include "Cubemap.h"

#include <stdexcept>
#include <algorithm>
#include <chrono>
#include <cstdlib>
#include <cstdint>
#include <limits>
#include <unordered_map>

const std::vector<const char*> validationLayers = {
    "VK_LAYER_KHRONOS_validation"
};

const std::vector<const char*> deviceExtensions = {
    VK_KHR_SWAPCHAIN_EXTENSION_NAME
};

#ifdef NDEBUG
const bool enableValidationLayers = false;
#else
const bool enableValidationLayers = true;
#endif


struct QueueFamilyIndices {
    std::optional<uint32_t> graphicsAndComputeFamily;
    std::optional<uint32_t> presentFamily;

    bool isComplete() {
        return graphicsAndComputeFamily.has_value() && presentFamily.has_value();
    }
};

struct SwapChainSupportDetails {
    VkSurfaceCapabilitiesKHR capabilities;
    std::vector<VkSurfaceFormatKHR> formats;
    std::vector<VkPresentModeKHR> presentModes;
};

/**
 *  A class that handles the main Vulkan parts of this program.
 */
class VulkanHandler {
public:

    VulkanHandler() {};
    VulkanHandler(GLFWwindow* window) : window(window) {};

    void initVulkan();
    void drawFrame(ObjectHandler& objHandler, Camera& camera, float t, float dt, Cubemap& cubemap, Cubemap& dynEnvironmentMap, ParticleObject& particleSystem, DefferedDrawObject& defferedPlane, DrawBuffer drawFlag);
    void drawCubemapTexture(VkCommandBuffer commandBuffer, std::vector<VkFramebuffer> framebuffer, ObjectHandler& objHandler, Cubemap& cubemap, Cubemap& dynEnvironmentMap, ParticleObject& particleSystem);
    void dispatchParticleSystemCompute(ParticleObject& particleSystem);
    void cleanup();
    void cleanupDynEnvMap();
    void createDynEnvRenderPass(Cubemap* dynEnvMap);
    void createDefferedFramebuffers(DefferedDrawObject* defferedObject);
    void cleanupDefferedFramebuffers();

    void setFramebufferResized();
    void setDeviceWaitIdle();

    VkDevice& getDevice();
    VkPhysicalDevice& getPhysicalDevice();
    VkRenderPass& getRenderPass();
    VkRenderPass& getDynEnvMapRenderPass();
    uint32_t getCurrentFrame();
    VkCommandPool& getCommandPool();
    VkQueue& getGraphicsQueue();
    VkRenderPass& getDefferedRenderpass();

private:

    GLFWwindow* window;

    VkInstance instance;
    VkDebugUtilsMessengerEXT debugMessenger;
    VkSurfaceKHR surface;

    VkPhysicalDevice physicalDevice = VK_NULL_HANDLE;
    VkDevice device;

    VkQueue graphicsQueue;
    VkQueue computeQueue;
    VkQueue presentQueue;

    VkSwapchainKHR swapChain;
    std::vector<VkImage> swapChainImages;
    VkFormat swapChainImageFormat;
    VkExtent2D swapChainExtent;
    std::vector<VkImageView> swapChainImageViews;
    std::vector<VkFramebuffer> swapChainFramebuffers;

    std::vector<VkFramebuffer> dynEnvMapFirstFramebuffers;
    std::vector<VkFramebuffer> dynEnvMapSecondFramebuffers;
    std::vector<std::vector<VkFramebuffer>> dynEnvFramebufferSets;

    std::vector<VkImage> defferdImage;
    std::vector<VkImageView> defferedImageViews;
    std::vector<VkFramebuffer> defferedFramebuffers;

    VkRenderPass renderPass;
    VkRenderPass dynEnvRenderpass;
    VkRenderPass defferedRenderpass;

    VkCommandPool commandPool;

    VkImage depthImage;
    VkDeviceMemory depthImageMemory;
    VkImageView depthImageView;

    VkImage dynEnvDepthImage;
    VkDeviceMemory dynEnvDepthImageMemory;
    VkImageView dynEnvDepthImageView;

    VkImage defferedDepthImage;
    VkDeviceMemory defferedDepthImageMemory;
    VkImageView defferedDepthImageView;

    std::vector<VkCommandBuffer> commandBuffers;
    std::vector<VkCommandBuffer> computeCommandBuffers;
    std::vector<VkCommandBuffer> dynEnvCommandBuffers;

    std::vector<VkSemaphore> imageAvailableSemaphores;
    std::vector<VkSemaphore> renderFinishedSemaphores;
    std::vector<VkSemaphore> computeFinishedSemaphores;
    std::vector<VkSemaphore> cubemapFinishedSemaphores;
    std::vector<VkSemaphore> defferedFinishedSemaphores;
    std::vector<VkFence> inFlightFences;
    std::vector<VkFence> computeInFlightFences;
    std::vector<VkFence> cubemapInFlightFences;
    std::vector<VkFence> defferedInFlightFences;
    uint32_t currentFrame = 0;

    bool framebufferResized = false;

    void cleanupSwapChain();

    void recreateSwapChain();

    void createInstance();

    void populateDebugMessengerCreateInfo(VkDebugUtilsMessengerCreateInfoEXT& createInfo);

    void createSurface();

    void pickPhysicalDevice();

    void createLogicalDevice();

    void createSwapChain();

    VkCommandBuffer beginSingleTimeCommands();

    void endSingleTimeCommands(VkCommandBuffer commandBuffer);

    void createImageViews();

    void createDynEnvDepthResources(Cubemap* dynEnvMap);

    void createRenderPass();

    void createFramebuffers();

    void createDynEnvFramebuffers(Cubemap* DynEnvMap);

    void recordComputeCommandBuffer(VkCommandBuffer commandBuffer, ParticleObject& particleSystem);

    void createCommandPool();

    void createComputeCommandBuffers();

    void createDynEnvCommandBuffers();

    void createDepthResources();

    void createDefferedRenderpass();

    VkFormat findSupportedFormat(const std::vector<VkFormat>& candidates, VkImageTiling tiling, VkFormatFeatureFlags features);

    VkFormat findDepthFormat();

    bool hasStencilComponent(VkFormat format);

    uint32_t findMemoryType(uint32_t typeFilter, VkMemoryPropertyFlags properties);

    void createCommandBuffers();

    void recordCommandBuffer(VkCommandBuffer commandBuffer, uint32_t imageIndex, ObjectHandler& objHandler, Cubemap& cubemap, Cubemap& dynEnvironmentMap, ParticleObject& particleSystem, DefferedDrawObject& defferedPlane, DrawBuffer drawFlag);

    void createSyncObjects();

    void drawDeffered(VkCommandBuffer commandBuffer, uint32_t imageIndex, ObjectHandler& objHandler, Cubemap& cubemap, Cubemap& dynEnvironmentMap, ParticleObject& particleSystem);

    VkImageView createImageView(VkImage image, VkFormat format, VkImageAspectFlags aspectFlags);

    void createImage(uint32_t width, uint32_t height, VkFormat format,
        VkImageTiling tiling, VkImageUsageFlags usage, VkMemoryPropertyFlags properties, VkImage& image, VkDeviceMemory& imageMemory);

    VkSurfaceFormatKHR chooseSwapSurfaceFormat(const std::vector<VkSurfaceFormatKHR>& availableFormats);

    VkPresentModeKHR chooseSwapPresentMode(const std::vector<VkPresentModeKHR>& availablePresentModes);

    VkExtent2D chooseSwapExtent(const VkSurfaceCapabilitiesKHR& capabilities);

    SwapChainSupportDetails querySwapChainSupport(VkPhysicalDevice device);

    bool isDeviceSuitable(VkPhysicalDevice device);

    bool checkDeviceExtensionSupport(VkPhysicalDevice device);

    QueueFamilyIndices findQueueFamilies(VkPhysicalDevice device);

    std::vector<const char*> getRequiredExtensions();

    bool checkValidationLayerSupport();

    void setupDebugMessenger();

    VkResult CreateDebugUtilsMessengerEXT(VkInstance instance, const VkDebugUtilsMessengerCreateInfoEXT* pCreateInfo, const VkAllocationCallbacks* pAllocator, VkDebugUtilsMessengerEXT* pDebugMessenger);

    std::vector<char> readFile(const std::string& filename);

    static VKAPI_ATTR VkBool32 VKAPI_CALL debugCallback(VkDebugUtilsMessageSeverityFlagBitsEXT messageSeverity, VkDebugUtilsMessageTypeFlagsEXT messageType, const VkDebugUtilsMessengerCallbackDataEXT* pCallbackData, void* pUserData);

    void DestroyDebugUtilsMessengerEXT(VkInstance instance, VkDebugUtilsMessengerEXT debugMessenger, const VkAllocationCallbacks* pAllocator);
};

