#version 450

#define PI 3.1415926538

layout(binding = 0) uniform LightBufferObject {
    vec3 dirLDirection;
    vec3 dirLColor;
    vec3 cameraWorldPos;
} lubo;

layout(binding = 1) uniform sampler2D albedoSampler;
layout(binding = 2) uniform sampler2D normalSampler;
layout(binding = 3) uniform sampler2D coefficientsSampler;
layout(binding = 4) uniform sampler2D materialSampler;
layout(binding = 5) uniform sampler2D worldPosSampler;

layout(push_constant) uniform FlagPush {
    int flag;
} push;

layout(location = 0) in vec2 texCoord;

layout(location = 0) out vec4 outColor;

// Fresnel-Schlick Approximation
vec3 fresnel(float HdotV, vec3 reflectivity){
    return reflectivity + (1.0 - reflectivity) * pow(1.0 - HdotV, 5.0);
}

// Geometry function for shadowing and masking
float geometricAttenuation(float NdotH, float NdotV, float NdotL, float HdotV) {
    float term1 = (2*NdotH*NdotV)/(HdotV);
    float term2 = (2*NdotH*NdotL)/(HdotV);
    float minTerm = min(term1, term2);
    return min(1.0, minTerm);
}

// GGX Distribution function for microfacet model
float distributionGGX(vec3 N, vec3 H, float a)
{
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;
    
    float nom    = a2;
    float denom  = (NdotH2 * (a2 - 1.0) + 1.0);
    denom        = PI * denom * denom;
    
    return nom / denom;
}

void main() {
    int outputFlag = push.flag;
    
    vec3 albedo = texture(albedoSampler, texCoord).rgb;
    vec3 normal = texture(normalSampler, texCoord).rgb;
    vec3 coefficients = texture(coefficientsSampler, texCoord).rgb;
    vec3 material = texture(materialSampler, texCoord).rgb;
    float metalic = material.x;
    float roughness = material.y;
    float specularAlpha = material.z;
    vec3 worldPos = texture(worldPosSampler, texCoord).rgb;
    vec3 lightDirection = normalize(lubo.dirLDirection);

    if (outputFlag == 0) {
        // Normalize normal vector to ensure correct lighting calculations
        vec3 N = normalize(normal.rgb * 2.0 - 1.0);

        // Calculate view direction (from fragment to camera)
        vec3 V = normalize(lubo.cameraWorldPos - worldPos);
        
        // Ambient term
        vec3 texColor = albedo;
        vec4 color = vec4(texColor * coefficients.x, 1.0f); // ambient
        
        // Reflectivity term (Fresnel-Schlick approximation)
        vec3 reflectivity = mix(vec3(0.03), texColor, metalic);

        // Directional light calculations
        vec3 L = normalize(lightDirection);  // Light direction
        vec3 H = normalize(V + L);          // Halfway vector
        vec3 radiance = lubo.dirLColor;     // Light color

        // Calculate dot products for BRDF calculations
        float HdotV = max(dot(H, V), 0.0000001); // Avoid division by zero
        float NdotH = max(dot(N, H), 0.0000001); // Avoid division by zero
        float NdotL = max(dot(N, L), 0.0000001); // Avoid division by zero
        float NdotV = max(dot(N, V), 0.0000001); // Avoid division by zero

        // Geometry (masking and shadowing) term using Smith's method
        float G = geometricAttenuation(NdotH, NdotV, NdotL, HdotV);

        // Distribution (microfacet model) term using GGX
        float D = distributionGGX(N, H, roughness);

        // Fresnel term (Schlick approximation)
        vec3 F = fresnel(HdotV, reflectivity);

        // Cook-Torrance BRDF specular term
        vec3 specular = D * G * F;
        specular /= (4.0 * NdotV * NdotL); // Normalize for energy conservation

        // Diffuse term: Cook-Torrance diffuse term (Lambertian) with Fresnel
        vec3 kD = vec3(1.0) - F; // F is the specular part, so 1 - F is the diffuse part
        kD *= 1.0 - metalic; // More metallic, less diffuse

        // Ensure the diffuse term is calculated correctly
        color.rgb += (kD * texColor + specular) * max(0.0, NdotL) * radiance;
		
        outColor = color;
		
    } else if (outputFlag == 1) {
        outColor = vec4(albedo, 1.0f); // Output albedo color
    } else if (outputFlag == 2) {
        outColor = vec4(normal, 1.0f); // Output normal
    } else if (outputFlag == 3) {
        outColor = vec4(coefficients, 1.0f); // Output coefficients (like ambient)
    } else if (outputFlag == 4) {
        outColor = vec4(metalic); // Output metallic value
    } else if (outputFlag == 5) {
        outColor = vec4(roughness); // Output roughness value
    } else if (outputFlag == 6) {
        outColor = vec4(specularAlpha); // Output specular alpha value
    } else if (outputFlag == 7) {
        outColor = vec4(texture(worldPosSampler, texCoord).rgb, 1.0f); // Output world position
    } else {
        outColor = vec4(1.0f); // Default output if no flag matched
    }
}
